#//
#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc. 
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
"""
Title: Policy Classes

Policy classes are used to implement polymorphic operations that
differ between built-in types and class-based types. Generic
components can then be built that work with either classes or 
built-in types, depending on what policy class is used.
"""

from ..base.sv import sv


#//----------------------------------------------------------------------
#// CLASS: UVMBuiltInComp #(T)
#// 
#// This policy class is used to compare built-in types.
#//
#// Provides a comp method that compares the built-in type,
#// T, for which the == operator is defined.
#//----------------------------------------------------------------------

class UVMBuiltInComp:  # (type T=int)

    @classmethod
    def comp(self, a, b):
        return a == b



#//----------------------------------------------------------------------
#// CLASS: UVMBuiltInConverter #(T)
#//
#// This policy class is used to convert built-in types to strings.
#//
#// Provides a convert2string method that converts the built-in type, T,
#// to a string using the %p format specifier.
#//----------------------------------------------------------------------


class UVMBuiltInConverter:  # (type T=int)
    @classmethod
    def convert2string(self, t):
        return sv.sformatf("%p", t)




#//----------------------------------------------------------------------
#// CLASS: UVMBuiltInClone #(T)
#//
#// This policy class is used to clone built-in types via the = operator.
#//
#// Provides a clone method that returns a copy of the built-in type, T.
#//----------------------------------------------------------------------


class UVMBuiltInClone:  # (type T=int)
    @classmethod
    def clone(self, from_):
        return from_


#//----------------------------------------------------------------------
#// CLASS: UVMClassComp #(T)
#//
#// This policy class is used to compare two objects of the same type.
#//
#// Provides a comp method that compares two objects of type T. The
#// class T must provide the method "function bit compare(T rhs)",
#// similar to the <uvm_object::compare> method.
#//----------------------------------------------------------------------

class UVMClassComp:  # (type T=int)
    @classmethod
    def comp(self, a, b):
        return a.compare(b)


#//----------------------------------------------------------------------
#// CLASS: UVMClassConverter #(T)
#//
#// This policy class is used to convert a class object to a string.
#//
#// Provides a convert2string method that converts an instance of type T
#// to a string. The class T must provide the method
#// "function string convert2string()",
#// similar to the <uvm_object::convert2string> method.
#//----------------------------------------------------------------------

class UVMClassConverter:  # (type T=int)

    @classmethod
    def convert2string(self, t):
        return t.convert2string()


#//----------------------------------------------------------------------
#// CLASS: UVMClassClone #(T)
#//
#// This policy class is used to clone class objects.
#//
#// Provides a clone method that returns a copy of the built-in type, T.
#// The class T must implement the clone method, to which this class
#// delegates the operation. If T is derived from <uvm_object>, then
#// T must instead implement <uvm_object::do_copy>, either directly or
#// indirectly through use of the `uvm_field macros.
#//----------------------------------------------------------------------

class UVMClassClone:  # (type T=int)
    @classmethod
    def clone(self, from_):
        return from_.clone()



#-----------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#-----------------------------------------------------------------------------

from ..base.uvm_debug import uvm_debug
from ..base.uvm_globals import UVM_NONE, uvm_report_info
from ..base.sv import sv

#//-----------------------------------------------------------------------------
#// Title: Callback Macros
#//
#// These macros are used to register and execute callbacks extending
#// from ~uvm_callbacks~.
#//-----------------------------------------------------------------------------


#//-----------------------------------------------------------------------------
#// MACRO: `uvm_register_cb
#//
#//| `uvm_register_cb(T,CB)
#//
#// Registers the given ~CB~ callback type with the given ~T~ object type. If
#// a type-callback pair is not registered then a warning is issued if an
#// attempt is made to use the pair (add, delete, etc.).
#//
#// The registration will typically occur in the component that executes the
#// given type of callback. For instance:
#//
#//| virtual class mycb extends uvm_callback;
#//|   virtual function void doit();
#//| endclass
#//|
#//| class my_comp extends uvm_component;
#//|   `uvm_register_cb(my_comp,mycb)
#//|   ...
#//|   task run_phase(uvm_phase phase);
#//|     ...
#//|     `uvm_do_callbacks(my_comp, mycb, doit())
#//|   endtask
#//| endclass
#//-----------------------------------------------------------------------------

#`define uvm_register_cb(T,CB) \
#  static local bit m_register_cb_``CB = uvm_callbacks#(T,CB)::m_register_pair(`"T`",`"CB`");
def uvm_register_cb(T, CB):
    from ..base.uvm_callback import UVMCallbacks
    Ts = T.__name__
    cb_name = "_m_register_cb_" + CB.__name__
    ok = UVMCallbacks.m_register_pair(Ts, CB.__name__, T)
    setattr(T, cb_name, ok)


#//-----------------------------------------------------------------------------
#// MACRO: `uvm_set_super_type
#//
#//| `uvm_set_super_type(T,ST)
#//
#// Defines the super type of ~T~ to be ~ST~. This allows for derived class
#// objects to inherit typewide callbacks that are registered with the base
#// class.
#//
#// The registration will typically occur in the component that executes the
#// given type of callback. For instance:
#//
#//| virtual class mycb extend uvm_callback;
#//|   virtual function void doit();
#//| endclass
#//|
#//| class my_comp extends uvm_component;
#//|   `uvm_register_cb(my_comp,mycb)
#//|   ...
#//|   task run_phase(uvm_phase phase);
#//|     ...
#//|     `uvm_do_callbacks(my_comp, mycb, doit())
#//|   endtask
#//| endclass
#//|
#//| class my_derived_comp extends my_comp;
#//|   `uvm_set_super_type(my_derived_comp,my_comp)
#//|   ...
#//|   task run_phase(uvm_phase phase);
#//|     ...
#//|     `uvm_do_callbacks(my_comp, mycb, doit())
#//|   endtask
#//| endclass
#//-----------------------------------------------------------------------------

#`define uvm_set_super_type(T,ST) \
#  static local bit m_register_``T``ST = uvm_derived_callbacks#(T,ST)::register_super_type(`"T`",`"ST`");


#-----------------------------------------------------------------------------
# MACRO: `uvm_do_callbacks
#
#| `uvm_do_callbacks(T,CB,METHOD)
#
# Calls the given ~METHOD~ of all callbacks of type ~CB~ registered with
# the calling object (i.e. ~this~ object), which is or is based on type ~T~.
#
# This macro executes all of the callbacks associated with the calling
# object (i.e. ~this~ object). The macro takes three arguments:
#
# - CB is the class type of the callback objects to execute. The class
#   type must have a function signature that matches the METHOD argument.
#
# - T is the type associated with the callback. Typically, an instance
#   of type T is passed as one the arguments in the ~METHOD~ call.
#
# - METHOD is the method call to invoke, with all required arguments as
#   if they were invoked directly.
#
# For example, given the following callback class definition:
#
#| virtual class mycb extends uvm_cb;
#|   pure function void my_function (mycomp comp, int addr, int data);
#| endclass
#
# A component would invoke the macro as
#
#| task mycomp::run_phase(uvm_phase phase);
#|    int curr_addr, curr_data;
#|    ...
#|    `uvm_do_callbacks(mycb, mycomp, my_function(this, curr_addr, curr_data))
#|    ...
#| endtask
#-----------------------------------------------------------------------------

def uvm_do_callbacks(self, CB, METHOD, *args):
    uvm_debug(self, 'uvm_do_callbacks', 'Exec CBs with ' + METHOD)
    uvm_do_obj_callbacks(self, CB, METHOD, *args)

async def uvm_do_callbacks_async(self, CB, METHOD, *args):
    uvm_debug(self, 'uvm_do_callbacks', 'Exec CBs with ' + METHOD)
    await uvm_do_obj_callbacks_async(self, CB, METHOD, *args)

#-----------------------------------------------------------------------------
# MACRO: `uvm_do_obj_callbacks
#
#| `uvm_do_obj_callbacks(T,CB,OBJ,METHOD)
#
# Calls the given ~METHOD~ of all callbacks based on type ~CB~ registered with
# the given object, ~OBJ~, which is or is based on type ~T~.
#
# This macro is identical to <`uvm_do_callbacks> macro,
# but it has an additional ~OBJ~ argument to allow the specification of an
# external object to associate the callback with. For example, if the
# callbacks are being applied in a sequence, ~OBJ~ could be specified
# as the associated sequencer or parent sequence.
#
#|    ...
#|    `uvm_do_callbacks(mycb, mycomp, seqr, my_function(seqr, curr_addr, curr_data))
#|    ...
#-----------------------------------------------------------------------------


def uvm_do_obj_callbacks(OBJ, CB, METHOD, *args):
    from ..base.uvm_callback import UVMCallbackIter
    cb_iter = UVMCallbackIter(OBJ, CB)
    cb = cb_iter.first()

    while cb is not None:
        uvm_cb_trace_noobj(cb, (
            "Executing callback method '{}' for callback {} (CB) from {} (T)"
            .format(METHOD, cb.get_name(), OBJ.get_full_name())))
        m_to_call = getattr(cb, METHOD)
        m_to_call(*args)
        cb = cb_iter.next()

async def uvm_do_obj_callbacks_async(OBJ, CB, METHOD, *args):
    from ..base.uvm_callback import UVMCallbackIter
    cb_iter = UVMCallbackIter(OBJ, CB)
    cb = cb_iter.first()

    while cb is not None:
        uvm_cb_trace_noobj(cb, (
            "Executing callback method '{}' for callback {} (CB) from {} (T)"
            .format(METHOD, cb.get_name(), OBJ.get_full_name())))
        m_to_call = getattr(cb, METHOD)
        await m_to_call(*args)
        cb = cb_iter.next()

#//-----------------------------------------------------------------------------
#// MACRO: `uvm_do_callbacks_exit_on
#//
#//| `uvm_do_callbacks_exit_on(T,CB,METHOD,VAL)
#//
#// Calls the given ~METHOD~ of all callbacks of type ~CB~ registered with
#// the calling object (i.e. ~this~ object), which is or is based on type ~T~,
#// returning upon the first callback returning the bit value given by ~VAL~.
#//
#// This macro executes all of the callbacks associated with the calling
#// object (i.e. ~this~ object). The macro takes three arguments:
#//
#// - CB is the class type of the callback objects to execute. The class
#//   type must have a function signature that matches the METHOD argument.
#//
#// - T is the type associated with the callback. Typically, an instance
#//   of type T is passed as one the arguments in the ~METHOD~ call.
#//
#// - METHOD is the method call to invoke, with all required arguments as
#//   if they were invoked directly.
#//
#// - VAL, if 1, says return upon the first callback invocation that
#//   returns 1. If 0, says return upon the first callback invocation that
#//   returns 0.
#//
#// For example, given the following callback class definition:
#//
#//| virtual class mycb extends uvm_cb;
#//|   pure function bit drop_trans (mycomp comp, my_trans trans);
#//| endclass
#//
#// A component would invoke the macro as
#//
#//| task mycomp::run_phase(uvm_phase phase);
#//|    my_trans trans;
#//|    forever begin
#//|      get_port.get(trans);
#//|      if(do_callbacks(trans) == 0)
#//|        uvm_report_info("DROPPED",{"trans dropped: %s",trans.convert2string()});
#//|      else
#//|        // execute transaction
#//|    end
#//| endtask
#//| function bit do_callbacks(my_trans);
#//|   // Returns 0 if drop happens and 1 otherwise
#//|   `uvm_do_callbacks_exit_on(mycomp, mycb, extobj, drop_trans(this,trans), 1)
#//| endfunction
#//
#// Because this macro calls ~return~, its use is restricted to implementations
#// of functions that return a ~bit~ value, as in the above example.
#//
#//-----------------------------------------------------------------------------


#`define uvm_do_callbacks_exit_on(T,CB,METHOD,VAL) \
#  `uvm_do_obj_callbacks_exit_on(T,CB,this,METHOD,VAL) \
def uvm_do_callbacks_exit_on(self,CB,METHOD,VAL, *args):
    uvm_do_obj_callbacks_exit_on(self, CB, METHOD, VAL, *args)


#//-----------------------------------------------------------------------------
#// MACRO: `uvm_do_obj_callbacks_exit_on
#//
#//| `uvm_do_obj_callbacks_exit_on(T,CB,OBJ,METHOD,VAL)
#//
#// Calls the given ~METHOD~ of all callbacks of type ~CB~ registered with
#// the given object ~OBJ~, which must be or be based on type ~T~, and returns
#// upon the first callback that returns the bit value given by ~VAL~. It is
#// exactly the same as the <`uvm_do_callbacks_exit_on> but has a specific
#// object instance (instead of the implicit this instance) as the third
#// argument.
#//
#//| ...
#//|  // Exit if a callback returns a 1
#//|  `uvm_do_callbacks_exit_on(mycomp, mycb, seqr, drop_trans(seqr,trans), 1)
#//| ...
#//
#// Because this macro calls ~return~, its use is restricted to implementations
#// of functions that return a ~bit~ value, as in the above example.
#//-----------------------------------------------------------------------------

#`define uvm_do_obj_callbacks_exit_on(T,CB,OBJ,METHOD,VAL) \
#   begin \
#     uvm_callback_iter#(T,CB) iter = new(OBJ); \
#     CB cb = iter.first(); \
#     while(cb != null) begin \
#       if (cb.METHOD == VAL) begin \
#         `uvm_cb_trace_noobj(cb,$sformatf(`"Executed callback method 'METHOD' for callback %s (CB) from %s (T) : returned value VAL (other callbacks will be ignored)`",cb.get_name(), OBJ.get_full_name())) \
#         return VAL; \
#       end \
#       `uvm_cb_trace_noobj(cb,$sformatf(`"Executed callback method 'METHOD' for callback %s (CB) from %s (T) : did not return value VAL`",cb.get_name(), OBJ.get_full_name())) \
#       cb = iter.next(); \
#     end \
#     return 1-VAL; \
#   end
def uvm_do_obj_callbacks_exit_on(OBJ, CB, METHOD, VAL, *args):
    from ..base.uvm_callback import UVMCallbackIter
    cb_iter = UVMCallbackIter(OBJ, CB)
    cb = cb_iter.first()

    while cb is not None:
        m_to_call = getattr(cb, METHOD)
        ret_val = m_to_call(*args)
        if ret_val == VAL:
            uvm_cb_trace_noobj(cb, sv.sformatf("Executed callback method "
                + "'METHOD' for callback %s (CB) from %s (T) : returned value VAL"
                + "(other callbacks will be ignored)l", cb.get_name(), OBJ.get_full_name()))
            return VAL
        uvm_cb_trace_noobj(cb, (
            "Executed callback method '{}' for callback {} (CB) from {} (T)"
            .format(METHOD, cb.get_name(), OBJ.get_full_name())))
        cb = cb_iter.next()
    return 1-VAL


#// The +define+UVM_CB_TRACE_ON setting will instrument the uvm library to emit
#// messages with message id UVMCB_TRC and UVM_NONE verbosity
#// notifing add,delete and execution of uvm callbacks. The instrumentation is off by default.
#

#`define uvm_cb_trace(OBJ,CB,OPER) \
#  begin \
#    string msg; \
#    msg = (OBJ == null) ? "null" : $sformatf("%s (%s@%0d)", \
#      OBJ.get_full_name(), OBJ.get_type_name(), OBJ.get_inst_id()); \
#    `uvm_info("UVMCB_TRC", $sformatf("%s: callback %s (%s@%0d) : to object %s",  \
#       OPER, CB.get_name(), CB.get_type_name(), CB.get_inst_id(), msg), UVM_NONE) \
#  end
#

#`define uvm_cb_trace_noobj(CB,OPER) \
#  begin \
#    if(uvm_callbacks_base::m_tracing) \
#      `uvm_info("UVMCB_TRC", $sformatf("%s : callback %s (%s@%0d)" ,  \
#       OPER, CB.get_name(), CB.get_type_name(), CB.get_inst_id()), UVM_NONE) \
#  end
def uvm_cb_trace_noobj(CB, OPER):
    from ..base.uvm_callback import UVMCallbacksBase
    if UVMCallbacksBase.m_tracing:
        uvm_report_info("UVMCB_TRC", ("{} : callback {} ({}@{})".format(
            OPER, CB.get_name(), CB.get_type_name(), CB.get_inst_id())), UVM_NONE)



#//------------------------------------------------------------------------------
#//   Copyright 2010-2011 Mentor Graphics Corporation
#//   All Rights Reserved Worldwide 
#//  
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------
#
#
#`ifndef UVM_NO_DEPRECATED
#
#//-----------------------------------------------------------------------------
#//
#// *** DEPRECATED ***
#// Group- Sequence Registration Macros
#//
#// The sequence-specific macros perform the same function as the set of
#// `uvm_object_*_utils macros, except they also set the default sequencer type
#// the sequence will run on. 
#//-----------------------------------------------------------------------------
#
#`define m_uvm_register_sequence(TYPE_NAME, SEQUENCER) \
#  static bit is_registered_with_sequencer = SEQUENCER``::add_typewide_sequence(`"TYPE_NAME`");
#
#
#// MACRO- `uvm_sequence_utils_begin
#//
#`define uvm_sequence_utils_begin(TYPE_NAME, SEQUENCER) \
#  `m_uvm_register_sequence(TYPE_NAME, SEQUENCER) \
#  `uvm_declare_p_sequencer(SEQUENCER) \
#  `uvm_object_utils_begin(TYPE_NAME)
#
#// MACRO- `uvm_sequence_utils_end
#//
#`define uvm_sequence_utils_end \
#  `uvm_object_utils_end
#
#// MACRO- `uvm_sequence_utils
#//
#// The sequence macros can be used in non-parameterized <uvm_sequence #(REQ,RSP)>
#// extensions to pre-register the sequence with a given <uvm_sequencer #(REQ,RSP)>
#// type.
#//
#// For sequences that do not use any `uvm_field macros:
#//
#//|  `uvm_sequence_utils(TYPE_NAME,SQR_TYPE_NAME)
#//
#// For sequences employing with field macros:
#//
#//|  `uvm_sequence_utils_begin(TYPE_NAME,SQR_TYPE_NAME)
#//|    `uvm_field_* macro invocations here
#//|  `uvm_sequence_utils_end
#//
#// The sequence-specific macros perform the same function as the set of
#// `uvm_object_*_utils macros except that they also register the sequence's
#// type, TYPE_NAME, with the given sequencer type, SQR_TYPE_NAME, and define
#// the p_sequencer variable and m_set_p_sequencer method.
#//
#// Use `uvm_sequence_utils[_begin] for non-parameterized classes and
#// `uvm_sequence_param_utils[_begin] for parameterized classes.
#
#`define uvm_sequence_utils(TYPE_NAME, SEQUENCER) \
#  `uvm_sequence_utils_begin(TYPE_NAME,SEQUENCER) \
#  `uvm_sequence_utils_end
#
#
#//-----------------------------------------------------------------------------
#//
#// *** DEPRECATED ***
#//
#// Group- Sequencer Registration Macros
#//
#// The sequencer-specific macros perform the same function as the set of
#// `uvm_componenent_*utils macros except that they also declare the plumbing
#// necessary for creating the sequencer's sequence library.
#//-----------------------------------------------------------------------------
#
#`define uvm_declare_sequence_lib \
#  protected bit m_set_sequences_called = 1;    \
#  static protected string m_static_sequences[$]; \
#  static protected string m_static_remove_sequences[$]; \
#  \
#  static function bit add_typewide_sequence(string type_name); \
#    m_static_sequences.push_back(type_name); \
#    return 1; \
#  endfunction\
#  \
#  static function bit remove_typewide_sequence(string type_name); \
#    m_static_remove_sequences.push_back(type_name); \
#    for (int i = 0; i < m_static_sequences.size(); i++) begin \
#      if (m_static_sequences[i] == type_name) \
#        m_static_sequences.delete(i); \
#    end \
#    return 1;\
#  endfunction\
#  \
#  function void uvm_update_sequence_lib();\
#    if(this.m_set_sequences_called) begin \
#      set_sequences_queue(m_static_sequences); \
#      this.m_set_sequences_called = 0;\
#    end\
#    for (int i = 0; i < m_static_remove_sequences.size(); i++) begin \
#      remove_sequence(m_static_remove_sequences[i]); \
#    end \
#  endfunction\
#
#
#
#// MACRO- `uvm_update_sequence_lib
#//
#// This macro populates the instance-specific sequence library for a sequencer.
#// It should be invoked inside the sequencer��s constructor.
#
#`define uvm_update_sequence_lib \
#  m_add_builtin_seqs(0); \
#  uvm_update_sequence_lib();
#
#
#// MACRO- `uvm_update_sequence_lib_and_item
#//
#// This macro populates the instance specific sequence library for a sequencer,
#// and it registers the given ~USER_ITEM~ as an instance override for the simple
#// sequence's item variable.
#//
#// The macro should be invoked inside the sequencer's constructor.
#
#`define uvm_update_sequence_lib_and_item(USER_ITEM) \
#  begin   uvm_coreservice_t cs = uvm_coreservice_t::get(); uvm_factory factory=cs.get_factory(); \
#  factory.set_inst_override_by_type( \
#    uvm_sequence_item::get_type(), USER_ITEM::get_type(), \
#  {get_full_name(), "*.item"}); end \
#  m_add_builtin_seqs(1); \
#  uvm_update_sequence_lib();
#
#
#// MACRO- `uvm_sequencer_utils
#
#`define uvm_sequencer_utils(TYPE_NAME) \
#  `uvm_sequencer_utils_begin(TYPE_NAME) \
#  `uvm_sequencer_utils_end
#
#// MACRO- `uvm_sequencer_utils_begin
#
#`define uvm_sequencer_utils_begin(TYPE_NAME) \
#  `uvm_declare_sequence_lib \
#  `uvm_component_utils_begin(TYPE_NAME)
#
#// MACRO- `uvm_sequencer_param_utils
#
#`define uvm_sequencer_param_utils(TYPE_NAME) \
#  `uvm_sequencer_param_utils_begin(TYPE_NAME) \
#  `uvm_sequencer_utils_end
#
#// MACRO- `uvm_sequencer_param_utils_begin
#
#`define uvm_sequencer_param_utils_begin(TYPE_NAME) \
#  `uvm_declare_sequence_lib \
#  `uvm_component_param_utils_begin(TYPE_NAME)
#
#
#// MACRO- `uvm_sequencer_utils_end
#//
#// The sequencer macros are used in uvm_sequencer-based class declarations
#// in one of four ways.
#//
#// For simple sequencers, no field macros
#//
#//   `uvm_sequencer_utils(SQR_TYPE_NAME)
#//
#// For simple sequencers, with field macros
#//
#//   `uvm_sequencer_utils_begin(SQR_TYPE_NAME)
#//     `uvm_field_* macros here
#//   `uvm_sequencer_utils_end
#//
#// For parameterized sequencers, no field macros
#//
#//   `uvm_sequencer_param_utils(SQR_TYPE_NAME)
#//
#// For parameterized sequencers, with field macros
#//
#//   `uvm_sequencer_param_utils_begin(SQR_TYPE_NAME)
#//     `uvm_field_* macros here
#//   `uvm_sequencer_utils_end
#//
#// The sequencer-specific macros perform the same function as the set of
#// `uvm_componenent_*utils macros except that they also declare the plumbing
#// necessary for creating the sequencer's sequence library. This includes:
#//
#// 1. Declaring the type-based static queue of strings registered on the
#//    sequencer type.
#//
#// 2. Declaring the static function to add strings to item #1 above.
#//
#// 3. Declaring the static function to remove strings to item #1 above.
#//
#// 4. Declaring the function to populate the instance specific sequence library
#//    for a sequencer.
#//
#// Use `uvm_sequencer_utils[_begin] for non-parameterized classes and
#// `uvm_sequencer_param_utils[_begin] for parameterized classes.
#
#`define uvm_sequencer_utils_end \
#  `uvm_component_utils_end
#
#
#
#//-----------------------------------------------------------------------------
#//
#// MACRO- `uvm_package
#//
#// Use `uvm_package to define the SV package and to create a bogus type to help 
#// automate triggering the static initializers of the package.
#// Use uvm_end_package to endpackage.
#//-----------------------------------------------------------------------------
#
#`define uvm_package(PKG) \
#  package PKG; \
#  class uvm_bogus_class extends uvm::uvm_sequence;\
#  endclass
#
#`define uvm_end_package \
#   endpackage
#
#
#//-----------------------------------------------------------------------------
#//
#// MACRO- `uvm_sequence_library_package
#//
#// This macro is used to trigger static initializers in packages. `uvm_package
#// creates a bogus type which gets referred to by uvm_sequence_library_package
#// to make a package-based variable of the bogus type.
#//-----------------------------------------------------------------------------
#
#`define uvm_sequence_library_package(PKG_NAME) \
#  import PKG_NAME``::*; \
#  PKG_NAME``::uvm_bogus_class M_``PKG_NAME``uvm_bogus_class
#
#`endif // UVM_NO_DEPRECATED




import cocotb
from cocotb.triggers import Timer
from uvm import (UVMTest, run_test, UVMDebug)
from uvm.macros import (uvm_component_utils, uvm_fatal)

from master_slave_pkg import env_top

test_dur = 1200  # NS

#UVMDebug.full_debug()

class master_slave_test(UVMTest):

    def __init__(self, name, parent):
        super().__init__(name, parent)

    def build_phase(self, phase):
        self.env = env_top("env_master_slave", self)

    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        await Timer(test_dur, "NS")
        phase.drop_objection(self)


    def check_phase(self, phase):
        if not self.env.all_ok():
            uvm_fatal("ENV_NOT_OK", "There were errors in the env")


uvm_component_utils(master_slave_test)


@cocotb.test()
async def master_slave_top(dut):
    await run_test()




import unittest
from uvm.base.sv import (sv, sv_obj, uvm_glob_to_re)


class Packet(sv_obj):
    def __init__(self, name):
        sv_obj.__init__(self)
        # self.rand("addr")
        self.addr = 0
        self.b_addr = 0
        self.tx_id = 0
        self.rand("addr", list(range(1, 4)))
        self.rand("b_addr", list(range(0, 256)))
        #self.constraint("addr", [1, 2, 3])
        #self.constraint("b_addr", range(0, 256))
        self.data = 0
        self.rand("data", [0, 1, 2, 4, 8, 16, 32])

        self.free_var = 0x0

        c1 = lambda addr, b_addr: b_addr == 3 * addr
        self.constraint(c1)


class TestSV(unittest.TestCase):

    def test_sv_obj(self):
        """ Test that basic object randomization works """
        pp = Packet("ethss")
        pp.tx_id = 888
        pp.data = 1234
        ok = pp.randomize()
        self.assertEqual(ok, True)

        self.assertTrue(pp.addr in [1, 2, 3])
        self.assertNotEqual(pp.data, 1234)
        self.assertEqual(pp.tx_id, 888)
        self.assertEqual(3 * pp.addr, pp.b_addr)

        #ok = pp.randomize_with(lambda addr: addr in [5, 6, 7])
        #self.assertEqual(ok, True)
        #self.assertEqual(pp.addr, 3 * pp.b_addr)

    def test_randomize_with(self):
        pp = Packet("ethss")
        ok = pp.randomize_with(lambda addr: addr in [1, 2])
        self.assertTrue(ok)
        self.assertTrue(pp.addr in [1, 2])

        # Check that previous constraint does nothing
        ok = pp.randomize_with(lambda addr: addr in [3, 4])
        self.assertTrue(ok)
        self.assertTrue(pp.addr in [3, 4])
        # Add extra constraint outside class
        #p2 = Packet('pkt_dfess')
        #p2.rand('free_var', range(0, (1 << 32) - 1))
        #ok = p2.randomize_with(lambda free_var, addr, b_addr: free_var == (addr + b_addr))
        #self.assertTrue(ok)
        #self.assertEqual(p2.free_var, p2.addr + p2.b_addr)


    def test_uvm_glob_to_re(self):
        str1 = "uvm_*"
        res1 = uvm_glob_to_re(str1)
        self.assertEqual(res1, 'uvm_.*')
        str2 = "uvm_*xxx*yyy"
        res2 = uvm_glob_to_re(str2)
        self.assertEqual(res2, 'uvm_.*xxx.*yyy')
        str3 = "my_top.xxx*"
        res3 = uvm_glob_to_re(str3)
        self.assertEqual(res3, 'my_top\\.xxx.*')

        str4 = "__top__.master[0].slave*"
        res4 = uvm_glob_to_re(str4)
        self.assertEqual(res4, '__top__\\.master\\[0\\]\\.slave.*')

    def test_cast(self):
        my_int = 6
        arr = []
        ok = sv.cast(arr, my_int, int)
        self.assertEqual(ok, True)
        self.assertEqual(arr[0], 6)
        arr = []
        ok = sv.cast(arr, my_int, str)
        self.assertEqual(ok, False)
        self.assertEqual(len(arr), 0)

    def test_sformatf(self):
        str1 = sv.sformatf("Number: %0d, String: %s", 555, "xxx")
        self.assertRegex(str1, "Number: 555")
        self.assertRegex(str1, "String: xxx")
        test_str = [
            ["Hexadecimal: %5h", 123],
            ["Time value: %0t", 55555]
        ]
        for pair in test_str:
            str1 = sv.sformatf(pair[0], pair[1])



    def test_clog2(self):
        self.assertEqual(sv.clog2(0), 0)
        self.assertEqual(sv.clog2(3), 2)
        self.assertEqual(sv.clog2(5), 3)
        self.assertEqual(sv.clog2(16), 4)
        for i in range(0, 64):
            self.assertEqual(sv.clog2(1 << i), i)


    def test_sscanf(self):
        str_in = "ab12 23cd s"
        format_in = "%d %s"
        misc = 0
        num, my_str = sv.sscanf(str_in, format_in, misc)
        self.assertEqual(num, 12)
        self.assertEqual(my_str, "23cd")

        str_fails = "abc abc"
        format_in = "%d %d"
        res = sv.sscanf(str_fails, format_in, misc)
        self.assertEqual(res, [])

        num_str = "0x1234 5678 10104 -345.124"
        format_in = "%h %d %b %f"
        res = sv.sscanf(num_str, format_in, misc)
        num_hex, num_dec, num_bin, num_float = res
        self.assertEqual(num_hex, 0x1234)
        self.assertEqual(num_dec, 5678)
        self.assertEqual(num_bin, 0b1010)
        self.assertEqual(num_float, -345.124)


if __name__ == '__main__':
    unittest.main()




import unittest
from uvm.base.uvm_mailbox import UVMMailbox


class TestUVMMailbox(unittest.TestCase):

    def test_try_put(self):
        fifo = UVMMailbox(2)
        self.assertEqual(fifo.can_put(), True)
        self.assertEqual(fifo.can_get(), False)

    def test_try_get(self):
        fifo = UVMMailbox(2)
        self.assertEqual(fifo.can_get(), False)
        itemq = []
        self.assertEqual(fifo.try_get(itemq), False)
        self.assertEqual(fifo.try_put(234), True)
        self.assertEqual(fifo.try_get(itemq), True)
        self.assertEqual(len(itemq), 1)
        self.assertEqual(itemq[0], 234)

    def test_unbounded(self):
        fifo = UVMMailbox()
        self.assertEqual(fifo.can_put(), True)
        for i in range(0, 10):
            self.assertEqual(fifo.try_put(i), True)
            arr = []
            self.assertEqual(fifo.try_peek(arr), True)
            self.assertEqual(arr[0], 0)

        for i in range(0, 10):
            arr = []
            self.assertEqual(fifo.try_get(arr), True)
            self.assertEqual(arr[0], i)



if __name__ == '__main__':
    unittest.main()



#
#-------------------------------------------------------------
#   Copyright 2004-2009 Synopsys, Inc.
#   Copyright 2010-2011 Mentor Graphics Corporation
#   Copyright 2010 Cadence Design Systems, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#-------------------------------------------------------------

from ..macros import uvm_fatal
from ..base.uvm_resource_db import ResourceDbClassFactory, UVMResourceDb

#------------------------------------------------------------------------------
# TITLE: Global Declarations for the Register Layer
#------------------------------------------------------------------------------
#
# This section defines globally available types, enums, and utility classes.
#
#------------------------------------------------------------------------------

#-------------
# Group: Types
#-------------
#
# Type: uvm_reg_data_t
#
# 2-state data value with <`UVM_REG_DATA_WIDTH> bits
#
#typedef  bit unsigned [`UVM_REG_DATA_WIDTH-1:0]  uvm_reg_data_t

# Type: uvm_reg_data_logic_t
#
# 4-state data value with <`UVM_REG_DATA_WIDTH> bits
#
#typedef  logic unsigned [`UVM_REG_DATA_WIDTH-1:0]  uvm_reg_data_logic_t

# Type: uvm_reg_addr_t
#
# 2-state address value with <`UVM_REG_ADDR_WIDTH> bits
#
#typedef  bit unsigned [`UVM_REG_ADDR_WIDTH-1:0]  uvm_reg_addr_t

# Type: uvm_reg_addr_logic_t
#
# 4-state address value with <`UVM_REG_ADDR_WIDTH> bits
#
#typedef  logic unsigned [`UVM_REG_ADDR_WIDTH-1:0]  uvm_reg_addr_logic_t

# Type: uvm_reg_byte_en_t
#
# 2-state byte_enable value with <`UVM_REG_BYTENABLE_WIDTH> bits
#
#typedef  bit unsigned [`UVM_REG_BYTENABLE_WIDTH-1:0]  uvm_reg_byte_en_t

# Type: uvm_reg_cvr_t
#
# Coverage model value set with <`UVM_REG_CVR_WIDTH> bits.
#
# Symbolic values for individual coverage models are defined
# by the <uvm_coverage_model_e> type.
#
# The following bits in the set are assigned as follows
#
# 0-7     - UVM pre-defined coverage models
# 8-15    - Coverage models defined by EDA vendors,
#           implemented in a register model generator.
# 16-23   - User-defined coverage models
# 24..    - Reserved
#
#typedef  bit [`UVM_REG_CVR_WIDTH-1:0]  uvm_reg_cvr_t

# Type: uvm_hdl_path_slice
#
# Slice of an HDL path
#
# Struct that specifies the HDL variable that corresponds to all
# or a portion of a register.
#
# path    - Path to the HDL variable.
# offset  - Offset of the LSB in the register that this variable implements
# size    - Number of bits (toward the MSB) that this variable implements
#
# If the HDL variable implements all of the register, ~offset~ and ~size~
# are specified as -1. For example:
#|
#| r1.add_hdl_path('{ '{"r1", -1, -1} })
#|

#typedef struct {
class uvm_hdl_path_slice:
    def __init__(self):
        self.path = ""
        self.offset = -1
        self.size = -1
#} uvm_hdl_path_slice
#

#typedef uvm_resource_db#(uvm_reg_cvr_t) uvm_reg_cvr_rsrc_db
uvm_reg_cvr_rsrc_db = ResourceDbClassFactory('uvm_reg_cvr_rsrc_db',
    [], int)


#--------------------
# Group: Enumerations
#--------------------

def enum_val_from(val, enum_names):
    if val < len(enum_names):
        return enum_names[val]
    else:
        raise Exception("Val {} out of range: {}".format(str(val),
            str(enum_names)))


# Enum: uvm_status_e
#
# Return status for register operations
#
# UVM_IS_OK      - Operation completed successfully
# UVM_NOT_OK     - Operation completed with error
# UVM_HAS_X      - Operation completed successfully bit had unknown bits.
#
UVM_IS_OK = 0
UVM_NOT_OK = 1
UVM_HAS_X = 2
UVM_STATUS_NAMES = ["UVM_IS_OK", "UVM_NOT_OK", "UVM_HAS_X"]

def uvm_status_e(name):
    return enum_val_from(name, UVM_STATUS_NAMES)


# Enum: uvm_path_e
#
# Path used for register operation
#
# UVM_FRONTDOOR    - Use the front door
# UVM_BACKDOOR     - Use the back door
# UVM_PREDICT      - Operation derived from observations by a bus monitor via
#                    the <uvm_reg_predictor> class.
# UVM_DEFAULT_PATH - Operation specified by the context

UVM_FRONTDOOR = 0
UVM_BACKDOOR = 1
UVM_PREDICT = 2
UVM_DEFAULT_PATH = 3
UVM_PATH_NAMES = ["UVM_FRONTDOOR", "UVM_BACKDOOR", "UVM_PREDICT",
        "UVM_DEFAULT_PATH"]

def uvm_path_e(name):
    return enum_val_from(name, UVM_PATH_NAMES)

# Enum: uvm_check_e
#
# Read-only or read-and-check
#
# UVM_NO_CHECK   - Read only
# UVM_CHECK      - Read and check
#
#   typedef enum {
UVM_NO_CHECK = 0
UVM_CHECK = 1
#   } uvm_check_e

# Enum: uvm_endianness_e
#
# Specifies byte ordering
#
# UVM_NO_ENDIAN      - Byte ordering not applicable
# UVM_LITTLE_ENDIAN  - Least-significant bytes first in consecutive addresses
# UVM_BIG_ENDIAN     - Most-significant bytes first in consecutive addresses
# UVM_LITTLE_FIFO    - Least-significant bytes first at the same address
# UVM_BIG_FIFO       - Most-significant bytes first at the same address
#
#   typedef enum {
UVM_NO_ENDIAN = 0
UVM_LITTLE_ENDIAN = 1
UVM_BIG_ENDIAN = 2
UVM_LITTLE_FIFO = 3
UVM_BIG_FIFO = 4
#   } uvm_endianness_e

# Enum: uvm_elem_kind_e
#
# Type of element being read or written
#
# UVM_REG      - Register
# UVM_FIELD    - Field
# UVM_MEM      - Memory location
#
#   typedef enum {
UVM_REG = 0
UVM_FIELD = 1
UVM_MEM = 2
UVM_ELEMENT_KIND_NAMES = ["UVM_REG", "UVM_FIELD", "UVM_MEM"]
#   } uvm_elem_kind_e

# Enum: uvm_access_e
#
# Type of operation begin performed
#
# UVM_READ     - Read operation
# UVM_WRITE    - Write operation
#
#   typedef enum {
UVM_READ = 0
UVM_WRITE = 1
UVM_BURST_READ = 2
UVM_BURST_WRITE = 3
UVM_ACCESS_NAMES = ["UVM_READ", "UVM_WRITE", "UVM_BURST_READ", "UVM_BURST_WRITE"]
#   } uvm_access_e

# Enum: uvm_hier_e
#
# Whether to provide the requested information from a hierarchical context.
#
# UVM_NO_HIER - Provide info from the local context
# UVM_HIER    - Provide info based on the hierarchical context
#
#   typedef enum {
UVM_NO_HIER = 0
UVM_HIER = 1
#   } uvm_hier_e

# Enum: uvm_predict_e
#
# How the mirror is to be updated
#
# UVM_PREDICT_DIRECT  - Predicted value is as-is
# UVM_PREDICT_READ    - Predict based on the specified value having been read
# UVM_PREDICT_WRITE   - Predict based on the specified value having been written
#
#   typedef enum {
UVM_PREDICT_DIRECT = 0
UVM_PREDICT_READ = 1
UVM_PREDICT_WRITE = 2
#   } uvm_predict_e

# Enum: uvm_coverage_model_e
#
# Coverage models available or desired.
# Multiple models may be specified by bitwise OR'ing individual model identifiers.
#
# UVM_NO_COVERAGE      - None
# UVM_CVR_REG_BITS     - Individual register bits
# UVM_CVR_ADDR_MAP     - Individual register and memory addresses
# UVM_CVR_FIELD_VALS   - Field values
# UVM_CVR_ALL          - All coverage models
#
#typedef enum uvm_reg_cvr_t {
UVM_NO_COVERAGE      = 0x0000
UVM_CVR_REG_BITS     = 0x0001
UVM_CVR_ADDR_MAP     = 0x0002
UVM_CVR_FIELD_VALS   = 0x0004
UVM_CVR_ALL          = -1
#} uvm_coverage_model_e

# Enum: uvm_reg_mem_tests_e
#
# Select which pre-defined test sequence to execute.
#
# Multiple test sequences may be selected by bitwise OR'ing their
# respective symbolic values.
#
# UVM_DO_REG_HW_RESET      - Run <uvm_reg_hw_reset_seq>
# UVM_DO_REG_BIT_BASH      - Run <uvm_reg_bit_bash_seq>
# UVM_DO_REG_ACCESS        - Run <uvm_reg_access_seq>
# UVM_DO_MEM_ACCESS        - Run <uvm_mem_access_seq>
# UVM_DO_SHARED_ACCESS     - Run <uvm_reg_mem_shared_access_seq>
# UVM_DO_MEM_WALK          - Run <uvm_mem_walk_seq>
# UVM_DO_ALL_REG_MEM_TESTS - Run all of the above
#
# Test sequences, when selected, are executed in the
# order in which they are specified above.
#
#typedef enum bit [63:0] {
UVM_DO_REG_HW_RESET      = 0x0000000000000001
UVM_DO_REG_BIT_BASH      = 0x0000000000000002
UVM_DO_REG_ACCESS        = 0x0000000000000004
UVM_DO_MEM_ACCESS        = 0x0000000000000008
UVM_DO_SHARED_ACCESS     = 0x0000000000000010
UVM_DO_MEM_WALK          = 0x0000000000000020
UVM_DO_ALL_REG_MEM_TESTS = 0xffffffffffffffff
#} uvm_reg_mem_tests_e
#

#-----------------------
# Group: Utility Classes
#-----------------------

#------------------------------------------------------------------------------
# Class: uvm_hdl_path_concat
#
# Concatenation of HDL variables
#
# A dArray of <uvm_hdl_path_slice> specifying a concatenation
# of HDL variables that implement a register in the HDL.
#
# Slices must be specified in most-to-least significant order.
# Slices must not overlap. Gaps may exist in the concatenation
# if portions of the registers are not implemented.
#
# For example, the following register
#|
#|        1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0
#| Bits:  5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
#|       +-+---+-------------+---+-------+
#|       |A|xxx|      B      |xxx|   C   |
#|       +-+---+-------------+---+-------+
#|
#
# If the register is implemented using a single HDL variable,
# The array should specify a single slice with its ~offset~ and ~size~
# specified as -1. For example:
#
#| concat.set('{ '{"r1", -1, -1} })
#
#------------------------------------------------------------------------------

class uvm_hdl_path_concat:

    def __init__(self):
        self.slices = []

    # Variable: slices
    # Array of individual slices,
    # stored in most-to-least significant order
    #uvm_hdl_path_slice slices[]

    # Function: set
    # Initialize the concatenation using an array literal
    def set(self, t):
        self.slices = t

    # Function: add_slice
    # Append the specified ~slice~ literal to the path concatenation
    def add_slice(self, slice):
        self.slices.append(slice)

    # Function: add_path
    # Append the specified ~path~ to the path concatenation,
    # for the specified number of bits at the specified ~offset~.
    def add_path(self, path, offset = -1, size = -1):
        t = uvm_hdl_path_slice()
        t.offset = offset
        t.path   = path
        t.size   = size
        self.add_slice(t)
    #endclass

# concat2string
#
def uvm_hdl_concat2string(concat):
    image = "{"

    if (len(concat.slices) == 1 and
        concat.slices[0].offset == -1 and
        concat.slices[0].size == -1):
        return concat.slices[0].path

    i = 0
    for sli in concat.slices:
        if i == 0:
            image = image + "" + sli.path
        else:
            image = image +  ", " + sli.path
        if sli.offset >= 0:
            image = image + "@" + "[{} +: {}]".format(sli.offset, sli.size)
        i += 1

    image = image + "}"
    return image


class UVMRegMapAddrRange:
    def __init__(self, _min=0, _max=0, stride=0):
        self.min = _min
        self.max = _max
        if isinstance(stride, int):
            self.stride = stride
        else:
            uvm_fatal("FLOAT_ERR", "stride must be integer. Got: " + str(stride))

    def convert2string(self):
        res = ("min: " + str(self.min) + ', max: ' + str(self.max) +
                ', stride: ' + str(self.stride))
        return res



def reg_test_off(model, test_patt):
    """ Reg test is disabled if given test_patt is found for that register """
    if isinstance(test_patt, list):
        res = False
        for patt in test_patt:
            res |= reg_test_off(model, patt)
        return res
    else:
        name = "REG::" + model.get_full_name()
        return (UVMResourceDb.get_by_name(name, test_patt, 0) is not None)


def reg_test_on(model, test_patt):
    return not reg_test_off(model, test_patt)



#from .uvm_barrier import *
from .uvm_bottomup_phase import *
from .uvm_callback import *
from .uvm_cmdline_processor import *
from .uvm_common_phases import *
from .uvm_component import *
from .uvm_config_db import *
from .uvm_coreservice import *
from .uvm_debug import *
from .uvm_domain import *
from .uvm_event import *
from .uvm_factory import *
from .uvm_global_vars import *
from .uvm_globals import *
from .uvm_mailbox import *
from .uvm_misc import *
from .uvm_object import *
from .uvm_object_globals import *
from .uvm_objection import *
from .uvm_phase import *
from .uvm_pool import *
from .uvm_port_base import *
from .uvm_printer import *
from .uvm_queue import *
from .uvm_recorder import *
from .uvm_registry import *
from .uvm_report_catcher import *
from .uvm_report_handler import *
from .uvm_report_message import *
from .uvm_report_object import *
from .uvm_report_server import *
from .uvm_resource import *
from .uvm_resource_db import *
from .uvm_root import *
from .uvm_runtime_phases import *
from .uvm_scope_stack import *
from .uvm_server import *
from .uvm_spell_chkr import *
from .uvm_task_phase import *
from .uvm_topdown_phase import *
from .uvm_transaction import *
from .sv import *



#
#------------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License or the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------


from .uvm_scope_stack import UVMScopeStack
from .uvm_object_globals import (UVM_BIN, UVM_DEC, UVM_HEX, UVM_NORADIX, UVM_OCT, UVM_STRING,
                                 UVM_TIME, UVM_UNSIGNED)
from .uvm_misc import (uvm_leaf_scope, uvm_integral_to_string,
        uvm_bitstream_to_string, uvm_object_value_str)
from .uvm_globals import uvm_report_error

UVM_STDOUT = 1  # Writes to standard out and logfile


class UVMPrinterRowInfo:
    def __init__(self):
        self.level = 0
        self.name = ""
        self.type_name = ""
        self.size = ""
        self.val = ""

    def __str__(self):
        return "L: {} Name: {} Type: {} Size: {} Val: {}".format(
            self.level, self.name, self.type_name, self.size, self.val)


class UVMPrinter():
    """
    Class: UVMPrinter

    The UVMPrinter class provides an interface for printing <uvm_objects> in
    various formats. Subtypes of UVMPrinter implement different print formats,
    or policies.

    A user-defined printer format can be created, or one of the following four
    built-in printers can be used:

    - `UVMPrinter` - provides base printer functionality; must be overridden.

    - `UVMTablePrinter` - prints the object in a tabular form.

    - `UVMTreePrinter` - prints the object in a tree form.

    - `UVMLinePrinter` - prints the information on a single line, but uses the
      same object separators as the tree printer.

    Printers have knobs that you use to control what and how information is printed.
    These knobs are contained in a separate knob class:

    - <uvm_printer_knobs> - common printer settings

    For convenience, global instances of each printer type are available for
    direct reference in your testbenches.

     -  <uvm_default_tree_printer>
     -  <uvm_default_line_printer>
     -  <uvm_default_table_printer>
     -  <uvm_default_printer> (set to default_table_printer by default)

    When <uvm_object::print> and <uvm_object::sprint> are called without
    specifying a printer, the <uvm_default_printer> is used.
    """


    def __init__(self, name=""):
        self.knobs = UVMPrinterKnobs()
        self.m_array_stack = []
        self.m_scope = UVMScopeStack()
        self.m_string = ""
        self.m_rows = []

    def emit(self):
        """
        emit
        ----
        Returns:
        """
        uvm_report_error("NO_OVERRIDE","emit() method not overridden in printer subtype")
        return ""

    def format_row(self, row):
        """
        format_row
        ----------
        Args:
            row:
        Returns:
        """
        return ""

    def format_header(self):
        """
        format_header
        ----------
        Returns:
        """
        return ""

    def format_footer(self):
        """
        format_footer
        ----------
        Returns:
        """
        return ""

    # print_array_header
    # ------------------

    def print_array_header(self, name, size, arraytype="array",
          scope_separator="."):
        row_info = UVMPrinterRowInfo()
        if name != "":
            self.m_scope.set_arg(name)

        row_info.level = self.m_scope.depth()
        row_info.name = self.adjust_name(self.m_scope.get(),scope_separator)
        row_info.type_name = arraytype
        row_info.size = "{}".format(size)
        row_info.val = "-"

        self.m_rows.append(row_info)

        self.m_scope.down(name)
        self.m_array_stack.append(1)


    # print_array_footer
    # ------------------

    def print_array_footer(self, size=0):
        if len(self.m_array_stack) > 0:
            self.m_scope.up()
            self.m_array_stack.pop()


    def print_array_range(self, min_val, max_val):
        """
        print_array_range
        -----------------
        Args:
            min_val:
            max_val:
        """
        tmpstr = ""
        if min_val == -1 and max_val == -1:
            return
        if min_val == -1:
            min_val = max_val
        if max_val == -1:
            max_val = min_val
        if max_val < min_val:
            return
        self.print_generic("...", "...", -2, "...")

    def print_object_header(self, name, value, scope_separator="."):
        """
        print_object_header
        -------------------
        Args:
            name:
            value:
            scope_separator:
        """
        row_info = UVMPrinterRowInfo()
        comp = None
        if name == "":
            if value is not None:
                if self.m_scope.depth() == 0:
                    comp = value
                    name = comp.get_full_name()
                else:
                    name = value.get_name()

        if name == "":
            name = "<unnamed>"

        self.m_scope.set_arg(name)
        row_info.level = self.m_scope.depth()

        if row_info.level == 0 and self.knobs.show_root is True:
            row_info.name = value.get_full_name()
        else:
            row_info.name = self.adjust_name(self.m_scope.get(),scope_separator)

        if value is None:
            row_info.type_name = "object"
        else:
            row_info.type_name = value.get_type_name()
        row_info.size = "-"

        if self.knobs.reference is True:
            row_info.val = uvm_object_value_str(value)
        else:
            row_info.val = "-"


        self.m_rows.append(row_info)

    def print_object(self, name, value, scope_separator="."):
        """
        print_object
        ------------
        Args:
            name:
            value:
            scope_separator:
        """
        comp = None
        child_comp = None
        self.print_object_header(name,value,scope_separator)

        if value is not None:
            knobs_ok = self.knobs.depth == -1 or self.knobs.depth > self.m_scope.depth()
            if knobs_ok and not (value in value._m_uvm_status_container.cycle_check):
                value._m_uvm_status_container.cycle_check[value] = 1
                if name == "" and value is not None:
                    self.m_scope.down(value.get_name())
                else:
                    self.m_scope.down(name)

                # Handle children of the comp
                #if($cast(comp, value)) begin
                comp = value
                name = ""
                if hasattr(comp, "has_first_child"):
                    if comp.has_first_child():
                        child_comp = comp.get_first_child()
                        ok = True
                        while ok is True:
                            if child_comp.print_enabled:
                                self.print_object("", child_comp)
                            ok = comp.has_next_child()
                            if ok is True:
                                child_comp = comp.get_next_child()

                # print members of object
                value.sprint(self)

                if name != "" and name[0] == "[":
                    self.m_scope.up("[")
                else:
                    self.m_scope.up(".")
                del value._m_uvm_status_container.cycle_check[value]

    def istop(self):
        """
        istop
        -----
        Returns:
        """
        return self.m_scope.depth() == 0

    def adjust_name(self, id, scope_separator="."):
        """
        #self.adjust_name
        -----------
        Args:
            id:
            scope_separator:
        Returns:
        """
        knobs_ok = self.knobs.show_root and self.m_scope.depth() == 0
        if knobs_ok or self.knobs.full_name or id == "...":
            return id
        return uvm_leaf_scope(id, scope_separator)

    # print_generic
    # -------------

    def print_generic(self, name, type_name, size, value, scope_separator="."):
        row_info = UVMPrinterRowInfo()

        if name != "" and name != "...":
            self.m_scope.set_arg(name)
            name = self.m_scope.get()

        row_info.level = self.m_scope.depth()
        row_info.name = self.adjust_name(name,scope_separator)
        row_info.type_name = type_name
        if size == -2:
            row_info.size = "..."
        else:
            row_info.size = "{}".format(size)

        if value == "":
            row_info.val = '""'
        else:
            row_info.val = value

        self.m_rows.append(row_info)

    def print_int(self, name, value, size, radix=UVM_NORADIX,
            scope_separator=".", type_name=""):
        self.print_field(name, value, size, radix, scope_separator, type_name)

    def print_field(self, name, value, size, radix=UVM_NORADIX,
            scope_separator=".", type_name=""):
        """
        print_field
        ---------
        Args:
            name:
            value:
            size:
            radix:
            UVM_NORADIX:
            scope_separator:
            type_name:
        """
        row_info = UVMPrinterRowInfo()
        sz_str = ""
        val_str = ""

        if name != "":
            self.m_scope.set_arg(name)
            name = self.m_scope.get()

        if type_name == "":
            if radix == UVM_TIME:
                type_name = "time"
            elif radix == UVM_STRING:
                type_name = "string"
            else:
                type_name = "integral"

        sz_str = "{}".format(size)

        if radix == UVM_NORADIX:
            radix = self.knobs.default_radix

        val_str = uvm_bitstream_to_string(value, size, radix,
                                       self.knobs.get_radix_str(radix))

        row_info.level = self.m_scope.depth()
        row_info.name = self.adjust_name(name,scope_separator)
        row_info.type_name = type_name
        row_info.size = sz_str
        row_info.val = val_str
        self.m_rows.append(row_info)


    # print_field_int
    # ---------

    def print_field_int(self, name, value, size, radix=UVM_NORADIX,
            scope_separator=".", type_name=""):

        row_info = UVMPrinterRowInfo()
        sz_str = ""
        val_str = ""

        if name != "":
            self.m_scope.set_arg(name)
            name = self.m_scope.get()

        if type_name == "":
            if radix == UVM_TIME:
                type_name = "time"
            elif radix == UVM_STRING:
                type_name = "string"
            else:
                type_name = "integral"

        sz_str = "{}".format(size)
        if radix == UVM_NORADIX:
            radix = self.knobs.default_radix

        val_str = uvm_integral_to_string(value, size, radix,
                                          self.knobs.get_radix_str(radix))

        row_info.level = self.m_scope.depth()
        row_info.name = self.adjust_name(name,scope_separator)
        row_info.type_name = type_name
        row_info.size = sz_str
        row_info.val = val_str

        self.m_rows.append(row_info)

    def print_time(self, name, value, scope_separator="."):
        """
        print_time
        ----------
        Args:
            name:
            value:
            scope_separator:
        """
        self.print_field_int(name, value, 64, UVM_TIME, scope_separator)

    # print_string
    # ------------

    def print_string(self, name, value, scope_separator="."):

        row_info = UVMPrinterRowInfo()
        if name != "":
            self.m_scope.set_arg(name)

        row_info.level = self.m_scope.depth()
        row_info.name = self.adjust_name(self.m_scope.get(),scope_separator)
        row_info.type_name = "string"
        row_info.size = "{}".format(len(value))
        if value == "":
            row_info.val = '""'
        else:
            row_info.val = value
        self.m_rows.append(row_info)

    # print_real
    # ----------

    def print_real(self, name, value, scope_separator="."):
        row_info = UVMPrinterRowInfo()
        if name != "" and name != "...":
            self.m_scope.set_arg(name)
            name = self.m_scope.get()

        row_info.level = self.m_scope.depth()
        row_info.name = self.adjust_name(self.m_scope.get(),scope_separator)
        row_info.type_name = "real"
        row_info.size = "64"
        row_info.val = "{}".format(value)
        self.m_rows.append(row_info)

    def index_string(self, index, name=""):
        """
        index_string
        ------------
        Args:
            index:
            name:
        """
        res = "{}".index
        res = name + "[" + res + "]"

#------------------------------------------------------------------------------
#
# Class: UVMPrinterKnobs
#
# The ~uvm_printer_knobs~ class defines the printer settings available to all
# printer subtypes.
#
#------------------------------------------------------------------------------
class UVMPrinterKnobs:

    def __init__(self):
        # Variable: header
        #
        # Indicates whether the <UVMPrinter::format_header> function should be called when
        # printing an object.
        self.header = True

        # Variable: footer
        #
        # Indicates whether the <UVMPrinter::format_footer> function should be called when
        # printing an object.
        self.footer = True

        # Variable: full_name
        #
        # Indicates whether <UVMPrinter::adjust_name> should print the full name of an identifier
        # or just the leaf name.
        self.full_name = False

        # Variable: identifier
        #
        # Indicates whether <UVMPrinter::adjust_name> should print the identifier. This is useful
        # in cases where you just want the values of an object, but no identifiers.
        self.identifier = True

        # Variable: type_name
        #
        # Controls whether to print a field's type name.
        self.type_name = True

        # Variable: size
        #
        # Controls whether to print a field's size.
        self.size = True

        # Variable: depth
        #
        # Indicates how deep to recurse when printing objects.
        # A depth of -1 means to print everything.
        self.depth = -1

        # Variable: reference
        #
        # Controls whether to print a unique reference ID for object handles.
        # The behavior of this knob is simulator-dependent.
        self.reference = True

        # Variable: begin_elements
        #
        # Defines the number of elements at the head of a list to print.
        # Use -1 for no max.
        self.begin_elements = 5

        # Variable: end_elements
        #
        # This defines the number of elements at the end of a list that
        # should be printed.
        self.end_elements = 5

        # Variable: prefix
        #
        # Specifies the string prepended to each output line
        self.prefix = ""

        # Variable: indent
        #
        # This knob specifies the number of spaces to use for level indentation.
        # The default level indentation is two spaces.
        self.indent = 2

        # Variable: show_root
        #
        # This setting indicates whether or not the initial object that is printed
        # (when current depth is 0) prints the full path name. By default, the first
        # object is treated like all other objects and only the leaf name is printed.
        self.show_root = False

        # Variable: mcd
        #
        # This is a file descriptor, or multi-channel descriptor, that specifies
        # where the print output should be directed.
        #
        # By default, the output goes to the standard output of the simulator.
        self.mcd = UVM_STDOUT

        # Variable: separator
        #
        # For tree printers only, determines the opening and closing
        # separators used for nested objects.
        self.separator = "{}"

        # Variable: show_radix
        #
        # Indicates whether the radix string ('h, and so on) should be prepended to
        # an integral value when one is printed.
        self.show_radix = True

        # Variable: default_radix
        #
        # This knob sets the default radix to use for integral values when no radix
        # enum is explicitly supplied to the <UVMPrinter::print_field> or
        # <UVMPrinter::print_field_int> methods.
        self.default_radix = UVM_HEX

        # Variable: dec_radix
        #
        # This string should be prepended to the value of an integral type when a
        # radix of <UVM_DEC> is used for the radix of the integral object.
        #
        # When a negative number is printed, the radix is not printed since only
        # signed decimal values can print as negative.
        self.dec_radix = "'d"

        # Variable: bin_radix
        #
        # This string should be prepended to the value of an integral type when a
        # radix of <UVM_BIN> is used for the radix of the integral object.
        self.bin_radix = "'b"

        # Variable: oct_radix
        #
        # This string should be prepended to the value of an integral type when a
        # radix of <UVM_OCT> is used for the radix of the integral object.
        self.oct_radix = "'o"

        # Variable: unsigned_radix
        #
        # This is the string which should be prepended to the value of an integral
        # type when a radix of <UVM_UNSIGNED> is used for the radix of the integral
        # object.
        self.unsigned_radix = "'d"

        # Variable: hex_radix
        #
        # This string should be prepended to the value of an integral type when a
        # radix of <UVM_HEX> is used for the radix of the integral object.
        self.hex_radix = "'h"

    def get_radix_str(self, radix):
        if self.show_radix is False:
            return ""
        if radix == UVM_NORADIX:
            radix = self.default_radix
        if radix == UVM_BIN:
            return self.bin_radix
        if radix == UVM_OCT:
            return self.oct_radix
        if radix == UVM_DEC:
            return self.dec_radix
        if radix == UVM_HEX:
            return self.hex_radix
        if radix == UVM_UNSIGNED:
            return self.unsigned_radix
        return ""



class UVMTablePrinter(UVMPrinter):
    """
    Class: UVMTablePrinter

    The table printer prints output in a tabular format.

    The following shows sample output from the table printer::

        ---------------------------------------------------
        Name        Type            Size        Value
        ---------------------------------------------------
        c1          container       -           @1013
        d1          mydata          -           @1022
        v1          integral        32          'hcb8f1c97
        e1          enum            32          THREE
        str         string          2           hi
        value       integral        12          'h2d
        ---------------------------------------------------

    """

    def __init__(self):
        UVMPrinter.__init__(self)
        self.m_max_name = 0
        self.m_max_type = 0
        self.m_max_size = 0
        self.m_max_value = 0

    def calculate_max_widths(self):
        self.m_max_name = 4
        self.m_max_type = 4
        self.m_max_size = 4
        self.m_max_value = 5
        for j in range(0, len(self.m_rows)):
            name_len = 0
            row = self.m_rows[j]
            name_len = self.knobs.indent*row.level + len(row.name)
            if name_len > self.m_max_name:
                self.m_max_name = name_len
            if len(row.type_name) > self.m_max_type:
                self.m_max_type = len(row.type_name)
            if len(row.size) > self.m_max_size:
                self.m_max_size = len(row.size)
            if len(row.val) > self.m_max_value:
                self.m_max_value = len(row.val)

    def emit(self):
        """
        Function: emit

        Formats the collected information from prior calls to ~print_*~
        into table format.

        Returns:
        """
        s = ""
        user_format = ""
        dash = ""
        space = ""
        dashes = ""
        linefeed = "\n" + self.knobs.prefix
        self.calculate_max_widths()

        q = [self. m_max_name, self.m_max_type,self.m_max_size,self.m_max_value,100]
        m = max(q)
        if len(dash) < m:
            dash = "-" * m
            space = " " * m

        if self.knobs.header is True:
            header = ""
            user_format = self.format_header()
            if user_format == "":
                dash_id = ""
                dash_typ = ""
                dash_sz = ""
                head_id = ""
                head_typ = ""
                head_sz = ""
                if self.knobs.identifier:
                    dashes = dash[1:self.m_max_name+2]
                    header = "Name" + space[1:self.m_max_name-2]
                if self.knobs.type_name is True:
                    dashes = dashes + dash[1:self.m_max_type+2]
                    header = header + "Type" + space[1:self.m_max_type-2]
                if self.knobs.size:
                    dashes = dashes + dash[1:self.m_max_size+2]
                    header = header + "Size" + space[1:self.m_max_size-2]
                dashes = dashes + dash[1:self.m_max_value] + linefeed
                header = header + "Value" + space[1:self.m_max_value-5] + linefeed

                s = s + dashes + header + dashes
            else:
                s = s + user_format + linefeed

        for i in range(0, len(self.m_rows)):
            row = self.m_rows[i]
            user_format = self.format_row(row)
            if user_format == "":
                row_str = ""
                if self.knobs.identifier:
                    row_str = (space[1:row.level * self.knobs.indent] + row.name
                        + space[1:self.m_max_name-len(row.name)-(row.level*self.knobs.indent)+2])
                if self.knobs.type_name:
                    row_str = (row_str + row.type_name
                        + space[1:(self.m_max_type-len(row.type_name)+2)])
                if self.knobs.size:
                    row_str = row_str + row.size + space[1:self.m_max_size-len(row.size)+2]
                s = s + row_str + row.val + space[1:self.m_max_value-len(row.val)] + linefeed
            else:
                s = s + user_format + linefeed

        if self.knobs.footer:
            user_format = self.format_footer()
            if user_format == "":
                s = s + dashes
            else:
                s = s + user_format + linefeed

        self.m_rows = []
        return self.knobs.prefix + s


class UVMTreePrinter(UVMPrinter):
    """
    Class: uvm_tree_printer

    By overriding various methods of the <UVMPrinter> super class,
    the tree printer prints output in a tree format.

    The following shows sample output from the tree printer::

      c1: (container@1013) {
        d1: (mydata@1022) {
             v1: 'hcb8f1c97
             e1: THREE
             str: hi
        }
        value: 'h2d
      }
    """


    def __init__(self):
        UVMPrinter.__init__(self)
        self.newline = "\n"
        self.knobs.size = 0
        self.knobs.type_name = 0
        self.knobs.header = 0
        self.knobs.footer = 0

    def emit(self):
        knobs = self.knobs
        s = knobs.prefix
        space = " " * 100
        user_format = ""
        newline = self.newline

        linefeed = newline + knobs.prefix
        if newline == "" or newline == " ":
            linefeed = newline

        # Header
        if knobs.header:
            user_format = self.format_header()
            if user_format != "":
                s = s + user_format + linefeed

        # Loops through all rows formatting them properly
        for i in range(len(self.m_rows)):
            row = self.m_rows[i]
            user_format = self.format_row(row)
            if user_format == "":
                indent_str = space[1:row.level * knobs.indent]

                # Name (id)
                if knobs.identifier:
                    s = s + indent_str + row.name
                    if row.name != "" and row.name != "...":
                        s = s + ": "

                # Type Name
                if row.val[0] == "@":  # is an object w/ knobs.reference on
                    s = s + "(" + row.type_name + row.val + ") "
                else:
                    if (knobs.type_name and
                         (row.type_name != "" or
                          row.type_name != "-" or
                          row.type_name != "...")):
                        s = s + "(" + row.type_name + ") "

                # Size
                if knobs.size:
                    if (row.size != "" or row.size != "-"):
                        s = s + "(" + row.size + ") "

                if i < (len(self.m_rows) - 1):
                    if self.m_rows[i+1].level > row.level:
                        sep = ""
                        if len(knobs.separator) > 0:
                            sep = str(knobs.separator[0])
                        s = s + sep + linefeed
                        continue

                # Value (unconditional)
                s = s + row.val + " " + linefeed

                # Scope handling...
                if i <= len(self.m_rows)-1:
                    end_level = 0
                    if (i == len(self.m_rows)-1):
                        end_level = 0
                    else:
                        end_level = self.m_rows[i+1].level
                    if (end_level < row.level):
                        indent_str = ""
                        for l in range(row.level-1, end_level-1, -1):
                            indent_str = space[1:l * knobs.indent+1]
                            sep = ""
                            if len(knobs.separator) > 1:
                                sep = str(knobs.separator[1])
                            s = s + indent_str + sep + linefeed
            else:
                s = s + user_format

        # Footer
        if knobs.footer:
            user_format = self.format_footer()
            if (user_format != ""):
                s = s + user_format + linefeed

        if (newline == "" or newline == " "):
            s = s + "\n"

        emit = s
        self.m_rows = []
        return emit

class UVMLinePrinter(UVMTreePrinter):
    """
    Class: UVMLinePrinter

    The line printer prints output in a line format.

    The following shows sample output from the line printer::
      c1: (container@1013) { d1: (mydata@1022) { v1: 'hcb8f1c97 e1: THREE str: hi } value: 'h2d }
    """

    def __init__(self):
        UVMTreePrinter.__init__(self)
        self.newline = " "
        self.knobs.indent = 0


# TODO: Print object info in JSON format
class UVMJSONPrinter(UVMPrinter):
    pass




import unittest
from uvm.base.uvm_report_object import UVMReportObject
from uvm.base.uvm_object_globals import (
    UVM_INFO, UVM_ERROR, UVM_LOG, UVM_DISPLAY, UVM_COUNT, UVM_MEDIUM, UVM_HIGH)


class TestUVMReportObject(unittest.TestCase):
    """ Unit tests for UVMReportObject """

    def test_verbosity(self):
        obj = UVMReportObject('rpt')
        rh = obj.get_report_handler()
        verb = rh.get_verbosity_level(UVM_INFO, "")
        self.assertEqual(verb, UVM_MEDIUM)
        self.assertEqual(obj.uvm_report_enabled(UVM_MEDIUM, UVM_INFO), True)

    def test_report_enabled(self):
        obj = UVMReportObject('rpt')
        is_en = obj.uvm_report_enabled(UVM_MEDIUM, UVM_INFO, id="")
        self.assertEqual(is_en, True)
        is_en = obj.uvm_report_enabled(UVM_HIGH, UVM_INFO, id="")
        self.assertEqual(is_en, False)

    def test_report_severity(self):
        obj = UVMReportObject('rpt')
        act = obj.get_report_action(UVM_ERROR, id="")
        self.assertEqual(act, UVM_DISPLAY | UVM_COUNT)
        act = obj.get_report_action(UVM_INFO, id="")
        self.assertEqual(act, UVM_DISPLAY)

        obj.set_report_severity_action(UVM_ERROR, UVM_LOG)
        act = obj.get_report_action(UVM_ERROR, id="")
        self.assertEqual(act, UVM_LOG)



if __name__ == '__main__':
    unittest.main()



#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer

from uvm.base import *
from uvm.comps import UVMTest
from uvm.macros import uvm_component_utils, uvm_fatal, uvm_info
from ubus_example_tb import ubus_example_tb
from ubus_example_master_seq_lib import read_modify_write_seq
from ubus_slave_seq_lib import slave_memory_seq
from uvm.seq import UVMSequence


class ubus_example_base_test(UVMTest):


    def __init__(self, name="ubus_example_base_test", parent=None):
        super().__init__(name, parent)
        self.test_pass = True
        self.ubus_example_tb0 = None
        self.printer = None

    def build_phase(self, phase):
        super().build_phase(phase)
        # Enable transaction recording for everything
        UVMConfigDb.set(self, "*", "recording_detail", UVM_FULL)
        # Create the tb
        self.ubus_example_tb0 = ubus_example_tb.type_id.create("ubus_example_tb0", self)
        # Create a specific depth printer for printing the created topology
        self.printer = UVMTablePrinter()
        self.printer.knobs.depth = 3

        arr = []
        if UVMConfigDb.get(None, "*", "vif", arr) is True:
            UVMConfigDb.set(self, "*", "vif", arr[0])
        else:
            uvm_fatal("NOVIF", "Could not get vif from config DB")


    def end_of_elaboration_phase(self, phase):
        # Set verbosity for the bus monitor for this demo
        if self.ubus_example_tb0.ubus0.bus_monitor is not None:
            self.ubus_example_tb0.ubus0.bus_monitor.set_report_verbosity_level(UVM_FULL)
        uvm_info(self.get_type_name(),
            sv.sformatf("Printing the test topology :\n%s", self.sprint(self.printer)), UVM_LOW)


    #  task run_phase(uvm_phase phase)
    #    //set a drain-time for the environment if desired
    #    phase.phase_done.set_drain_time(this, 50)
    #  endtask : run_phase

    def extract_phase(self, phase):
        self.err_msg = ""
        if self.ubus_example_tb0.scoreboard0.sbd_error:
            self.test_pass = False
            self.err_msg += '\nScoreboard error flag set'
        if self.ubus_example_tb0.scoreboard0.num_writes == 0:
            self.test_pass = False
            self.err_msg += '\nnum_writes == 0 in scb'
        if self.ubus_example_tb0.scoreboard0.num_init_reads == 0:
            self.test_pass = False
            self.err_msg += '\nnum_init_reads == 0 in scb'


    def report_phase(self, phase):
        if self.test_pass:
            uvm_info(self.get_type_name(), "** UVM TEST PASSED **", UVM_NONE)
        else:
            uvm_fatal(self.get_type_name(), "** UVM TEST FAIL **\n" +
                self.err_msg)


    #endclass : ubus_example_base_test
uvm_component_utils(ubus_example_base_test)

#// Read Modify Write Read Test

class test_read_modify_write(ubus_example_base_test):

    def __init__(self, name="test_read_modify_write", parent=None):
        super().__init__(name, parent)

    def build_phase(self, phase):
        UVMConfigDb.set(self, "ubus_example_tb0.ubus0.masters[0].sequencer.run_phase",
                "default_sequence", read_modify_write_seq.type_id.get())
        UVMConfigDb.set(self, "ubus_example_tb0.ubus0.slaves[0].sequencer.run_phase",
                "default_sequence", slave_memory_seq.type_id.get())
        #    // Create the tb
        super().build_phase(phase)

    
    async def run_phase(self, phase):
        phase.raise_objection(self, "test_read_modify_write OBJECTED")
        master_sqr = self.ubus_example_tb0.ubus0.masters[0].sequencer
        slave_sqr = self.ubus_example_tb0.ubus0.slaves[0].sequencer

        uvm_info("TEST_TOP", "Forking master_proc now", UVM_LOW)
        master_seq = read_modify_write_seq("r_mod_w_seq")
        master_proc = cocotb.fork(master_seq.start(master_sqr))

        slave_seq = slave_memory_seq("mem_seq")
        slave_proc = cocotb.fork(slave_seq.start(slave_sqr))
        #await [slave_proc, master_proc.join()]
        await sv.fork_join_any([slave_proc, master_proc])
        phase.drop_objection(self, "test_read_modify_write drop objection")
        if not master_seq.test_pass:
            self.test_pass = False


uvm_component_utils(test_read_modify_write)


#// Large word read/write test
#class test_r8_w8_r4_w4 extends ubus_example_base_test
#
#  `uvm_component_utils(test_r8_w8_r4_w4)
#
#  function new(string name = "test_r8_w8_r4_w4", uvm_component parent=null)
#    super.new(name,parent)
#  endfunction : new
#
#  virtual function void build_phase(uvm_phase phase)
#  begin
#    super.build_phase(phase)
#    uvm_config_db#(uvm_object_wrapper)::set(this,
#		      "ubus_example_tb0.ubus0.masters[0].sequencer.run_phase",
#			       "default_sequence",
#				r8_w8_r4_w4_seq::type_id::get())
#    uvm_config_db#(uvm_object_wrapper)::set(this,
#		      "ubus_example_tb0.ubus0.slaves[0].sequencer.run_phase",
#			       "default_sequence",
#				slave_memory_seq::type_id::get())
#  end
#  endfunction : build_phase
#
#endclass : test_r8_w8_r4_w4


# Virtual sequence for running masters and slaves
class seq_2m_4s(UVMSequence):
    def __init__(self, name):
        UVMSequence.__init__(self, name)
        # Populate these with sequencers to start sequences
        self.master_sqr = []
        self.slave_sqr = []

    
    async def body(self):
        # TODO create sequences and fork them
        i = 0
        for sqr in self.master_sqr:
            master_seq = read_modify_write_seq("r_mod_w_seq_" + str(i))
            master_proc = cocotb.fork(master_seq.start(sqr))
            i += 1
            # code...
        i = 0
        for sqr in self.slave_sqr:
            slave_seq = slave_memory_seq("mem_seq_" + str(i))
            slave_proc = cocotb.fork(slave_seq.start(sqr))
            i += 1
        await Timer(0, "NS")

#// 2 Master, 4 Slave test
class test_2m_4s(ubus_example_base_test):
    #
    #
    def __init__(self, name="test_2m_4s", parent=None):
        super().__init__(name, parent)

    def build_phase(self, phase):
        # loop_read_modify_write_seq lrmw_seq

        # Overides to the ubus_example_tb build_phase()
        # Set the topology to 2 masters, 4 slaves
        UVMConfigDb.set(self, "ubus_example_tb0.ubus0", "num_masters", 2)
        UVMConfigDb.set(self, "ubus_example_tb0.ubus0", "num_slaves", 4)
        ubus_example_base_test.build_phase(self, phase)

        # Control the number of RMW loops
        #    uvm_config_db#(int)::set(this,"ubus_example_tb0.ubus0.masters[0].sequencer.loop_read_modify_write_seq", "itr", 6)
        #    uvm_config_db#(int)::set(this,"ubus_example_tb0.ubus0.masters[1].sequencer.loop_read_modify_write_seq", "itr", 8)
        #
        #     // Define the sequences to run in the run phase
        #    uvm_config_db#(uvm_object_wrapper)::set(this,"*.ubus0.masters[0].sequencer.main_phase",
        #			       "default_sequence",
        #				loop_read_modify_write_seq::type_id::get())
        #     lrmw_seq = loop_read_modify_write_seq::type_id::create()
        #    uvm_config_db#(uvm_sequence_base)::set(this,
        #			       "ubus_example_tb0.ubus0.masters[1].sequencer.main_phase",
        #			       "default_sequence",
        #				lrmw_seq)
        #
        #     for(int i = 0; i < 4; i++):
        #	string slname
        #	$swrite(slname,"ubus_example_tb0.ubus0.slaves[%0d].sequencer", i)
        #	uvm_config_db#(uvm_object_wrapper)::set(this, {slname,".run_phase"},
        #						  "default_sequence",
        #						  slave_memory_seq::type_id::get())
        #     end
        #
        #    // Create the tb
        #    super.build_phase(phase)
        #  end
        #  endfunction : build_phase

    
    async def run_phase(self, phase):
        phase.raise_objection(self, "test_2m_4s objects", 1)
        top_seq = seq_2m_4s("seq_2m_4s")
        for slave in self.ubus_example_tb0.ubus0.slaves:
            top_seq.slave_sqr.append(slave.sequencer)
        for master in self.ubus_example_tb0.ubus0.masters:
            top_seq.master_sqr.append(master.sequencer)
        # TODO add handles for sequencers
        await top_seq.start(None)  # vseq runs with sequencer
        await Timer(1000, "NS")
        phase.drop_objection(self)


    def connect_phase(self, phase):
        # Connect other slaves monitor to scoreboard
        for i in range(1, 4):
            self.ubus_example_tb0.ubus0.slaves[i].monitor.item_collected_port.connect(
                self.ubus_example_tb0.scoreboard0.item_collected_export)


    def end_of_elaboration_phase(self, phase):
        # Set up slave address map for ubus0 (slaves[0] is overwritten here)
        self.ubus_example_tb0.ubus0.set_slave_address_map("slaves[0]", 0x0000, 0x3fff)
        self.ubus_example_tb0.ubus0.set_slave_address_map("slaves[1]", 0x4000, 0x7fff)
        self.ubus_example_tb0.ubus0.set_slave_address_map("slaves[2]", 0x8000, 0xBfff)
        self.ubus_example_tb0.ubus0.set_slave_address_map("slaves[3]", 0xC000, 0xFfff)
        super().end_of_elaboration_phase(phase)


uvm_component_utils(test_2m_4s)



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

#from typings import Dict
from uvm import (UVMComponent, UVMConfigDb, sv, uvm_get_array_index_string,
    uvm_is_match, uvm_error)


class ClassC(UVMComponent):


    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.v = 0  # type: int
        self.s = 0  # type: int
        self.myaa = {}  # Dict[str]


    def build_phase(self, phase):
        _str = []
        super().build_phase(phase)
        arr = []
        if UVMConfigDb.get(self, "", "v", arr):
            self.v = arr[0]
        arr = []
        if UVMConfigDb.get(self, "", "s", arr):
            self.s = arr[0]

        if UVMConfigDb.get(self, "", "myaa[foo]", _str):
            self.myaa["foo"] = _str[0]
            _str = []
        else:
            uvm_error("NO_CONF_MATCH", "Did not get myaa[foo]")

        if UVMConfigDb.get(self, "", "myaa[bar]", _str):
            self.myaa["bar"] = _str[0]
            _str = []
        else:
            uvm_error("NO_CONF_MATCH", "Did not get myaa[bar]")

        if UVMConfigDb.get(self, "", "myaa[foobar]", _str):
            self.myaa["foobar"] = _str[0]
            _str = []
        else:
            uvm_error("NO_CONF_MATCH", "Did not get myaa[foobar]")


    def get_type_name(self):
        return "C"

    def do_print(self, printer):
        printer.print_field("v", self.v, 32)
        printer.print_field("s", self.s, 32)
        printer.print_array_header("myaa", len(self.myaa), "aa_string_string")
        for key in self.myaa:
            printer.print_string(sv.sformatf("myaa[%0s]", key), self.myaa[key])
        printer.print_array_footer()


    def set_string_local(self, field_name, value, recurse=1):
        index = ""
        wildcarded = 0
        #call the super function to get child recursion and any registered fields
        super().set_string_local(field_name, value, recurse)

        index = uvm_get_array_index_string(field_name, wildcarded)
        if (not wildcarded and uvm_is_match(field_name, "myaa[" + index, "]")):
            self.myaa[index] = value




import cocotb
from cocotb.triggers import Timer, Join

from uvm_testlib import *

@cocotb.test(skip=False)
def test_tlm_nb_put_basic(dut):
    consumer = UVMTestConsumer('consumer', None)
    producer = UVMTestProducer('producer', None)

    producer.put_b_port.connect(consumer.put_b_imp)
    producer.get_b_port.connect(consumer.get_b_imp)
    producer.put_b_port.resolve_bindings()
    producer.get_b_port.resolve_bindings()

    producer.put_port.connect(consumer.put_imp)
    producer.get_port.connect(consumer.get_imp)
    producer.put_port.resolve_bindings()
    producer.get_port.resolve_bindings()

    if producer.put_b_port.m_if is None:
        raise Exception('put_b_port.m_if does not exist')

    pproc = cocotb.fork(producer.run_phase(None))
    #yield producer.run_phase(None)
    await consumer.wait_all_data()
    #yield [pproc.join(), wproc.join()]
    if consumer.count != 10:
        raise Exception('Consumer data count only {}'.format(consumer.count))

@cocotb.test(skip=False)
def test_tlm_exports(dut):
    producer = UVMTestProducer('producer', None)
    exporter = UVMTestExporter('exporter', None)
    exporter.connect_phase(None)
    producer.put_port.connect(exporter.put_export)
    producer.put_port.resolve_bindings()

    pproc = cocotb.fork(producer.run_phase(None))
    await exporter.wait_all_data()

@cocotb.test(skip=False)
def test_tlm_fifo(dut):
    fifo = UVMTLMAnalysisFIFO('fifo', None)
    fifo.write(1234);
    arr = []
    if not fifo.try_get(arr):
        raise Exception('FIFO try-get failed')





import unittest

from uvm.base.uvm_spell_chkr import UVMSpellChkr


class TestUVMSpellChkr(unittest.TestCase):

    def test_check_exact_match(self):
        strtab = {'xxx': 1, 'yyy': 2}
        self.assertTrue(UVMSpellChkr.check(strtab, 'xxx'))
        self.assertFalse(UVMSpellChkr.check(strtab, 'vvv'))


if __name__ == '__main__':
    unittest.main()



#//
#//------------------------------------------------------------------------------
#//   Copyright 2011 Mentor Graphics Corporation
#//   Copyright 2011 Cadence Design Systems, Inc.
#//   Copyright 2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------

from uvm.reg.uvm_reg_block import *
from uvm.reg.uvm_reg_model import *
from uvm.macros import *
from reg_B import reg_block_B


class reg_sys_S(UVMRegBlock):
    #
    # rand reg_block_B B[2]

    def __init__(self, name="S"):
        super().__init__(name, UVM_NO_COVERAGE)
        self.B = [None] * 2  # reg_block_B
        self.rand("B")
        #  endfunction: new

    def build(self):

        self.default_map = self.create_map("default_map", 0x0, 1, UVM_LITTLE_ENDIAN)

        for i in range(len(self.B)):
            self.B[i] = reg_block_B.type_id.create(sv.sformatf("B[%0d]", i))
            self.B[i].configure(self)
            self.B[i].build()
            self.default_map.add_submap(self.B[i].default_map, 0x100 + i * 0x100)


uvm_object_utils(reg_sys_S)




#//----------------------------------------------------------------------
#// Copyright 2010-2011 Mentor Graphics Corporation
#// Copyright 2010-2017 Synopsys, Inc.
#// Copyright 2010-2018 Cadence Design Systems, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#
#  `include "tlm2/uvm_tlm2_defines.svh"
#  `include "tlm2/uvm_tlm_time.svh"
#  `include "tlm2/uvm_tlm2_generic_payload.svh"
#  `include "tlm2/uvm_tlm2_ifs.svh"
#  `include "tlm2/uvm_tlm2_imps.svh"
#  `include "tlm2/uvm_tlm2_ports.svh"
#  `include "tlm2/uvm_tlm2_exports.svh"
#  `include "tlm2/uvm_tlm2_sockets_base.svh"
#  `include "tlm2/uvm_tlm2_sockets.svh"



#//
#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

#
# About: examples/uvm_tlm/bidir
# This example will illustrate how to create two threads one is the master and the other
# is the slave. The two threads are completely independent, each one of the two threads
# has two types of ports interfaces puts and gets.
#
# The communication between the two channel will be done through a uvm_tlm_req_rsp channel,
# which can handle the dual Request and Response of each thread
#
# This example will use The uvm_tlm_req_rsp channel, transaction-level ports, and the messaging
# facilities as part of the UVM library.

import cocotb
from uvm import *


#//----------------------------------------------------------------------
#// class master
#//----------------------------------------------------------------------

class master(UVMComponent):
    #
    #  uvm_blocking_put_port #(int) req_port
    #  uvm_blocking_get_port #(int) rsp_port
    #
    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.req_port = UVMBlockingPutPort("req_port", self)
        self.rsp_port = UVMBlockingGetPort("rsp_port", self)

    
    async def run_phase(self, phase):
        req_proc = cocotb.fork(self.request_process())
        rsp_proc = cocotb.fork(self.response_process())
        await sv.fork_join([req_proc, rsp_proc])
        #    fork
        #      request_process
        #      response_process
        #    join


    
    async def request_process(self):
        request_str = ""
        for i in range(10):
          request_str = sv.sformatf("%d", i)
          self.uvm_report_info("sending request   ", request_str, UVM_MEDIUM)
          await self.req_port.put(i)


    
    async def response_process(self):
        response = 0
        response_str = ""
        while True:
            arr = []
            await self.rsp_port.get(arr);   
            response = arr[0]
            response_str = sv.sformatf("%d", response)
            self.uvm_report_info("receiving response", response_str)



#//----------------------------------------------------------------------
#// class slave
#//----------------------------------------------------------------------

class slave(UVMComponent):


    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.req_port = UVMBlockingGetPort("req_port", self)
        self.rsp_port = UVMBlockingPutPort("rsp_port", self)

    
    async def run_phase(self, phase):
        request = 0
        response = 0
        request_str = ""
        response_str = ""

        while True:
            arr = []
            await self.req_port.get(arr)
            request = arr[0]
            request_str = sv.sformatf("%d", request)
            self.uvm_report_info("receiving request  ", request_str, UVM_MEDIUM)
            response = request
            response_str = sv.sformatf("%d", response)
            self.uvm_report_info("sending response   ", response_str, UVM_MEDIUM)
            await self.rsp_port.put(response)
           

#//----------------------------------------------------------------------
#// class bidir_env
#//----------------------------------------------------------------------

class bidir_env(UVMEnv):


    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.m = master("master", self)
        self.s = slave("slave", self)
        self.req_rsp = UVMTLMReqRspChannel("req_rsp_channel", self)


    def connect_phase(self, phase):
        self.m.req_port.connect(self.req_rsp.blocking_put_request_export)
        self.m.rsp_port.connect(self.req_rsp.blocking_get_response_export)
        self.s.req_port.connect(self.req_rsp.blocking_get_request_export)
        self.s.rsp_port.connect(self.req_rsp.blocking_put_response_export)


    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        #10
        await Timer(10, "NS")
        phase.drop_objection(self)


#//----------------------------------------------------------------------
#// module top
#//----------------------------------------------------------------------

@cocotb.test()
async def module_top(dut):
    env = bidir_env("env", None)
    uvm_info("TOP_ENV", "top-level env is " + env.get_name(), UVM_LOW)
    await run_test()




#// 
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#// 
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#// 
#//        http://www.apache.org/licenses/LICENSE-2.0
#// 
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//
#
#class hw_reset_test(test):
    #
    #   `uvm_component_utils(hw_reset_test)
    #
    #   def __init__(self, name, parent=None)
    #      super().__init__(name, parent)
    #   endfunction
    #
    #   local bit once = 1
    #async
    #   def main_phase(self, phase):
    #      if (once):
    #         once = 0
    #         phase.raise_objection(self)
    #         repeat (100 * 8) @(posedge env.vif.sclk)
    #         // This will clear the objection
    #         uvm_info("TEST", "Jumping back to reset phase", UVM_NONE)
    #         phase.jump(uvm_reset_phase::get())
    #      end
    #   endtask
    #   
from uvm.macros import *
    #endclass



#// 
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#// 
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#// 
#//        http://www.apache.org/licenses/LICENSE-2.0
#// 
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm import *


class reg_IntSrc(UVMReg):
    def __init__(self, name = "IntSrc"):
        super().__init__(name,9,UVM_NO_COVERAGE)
        self.TxEmpty = None
        self.TxLow = None
        self.TxFull = None
        self.RxEmpty = None
        self.RxHigh = None
        self.RxFull = None
        self.SA = None

    def build(self):
        self.TxEmpty = UVMRegField.type_id.create("TxEmpty",None, self.get_full_name())
        self.TxLow   = UVMRegField.type_id.create("TxLow",None, self.get_full_name())
        self.TxFull  = UVMRegField.type_id.create("TxFull",None, self.get_full_name())
        self.RxEmpty = UVMRegField.type_id.create("RxEmpty",None, self.get_full_name())
        self.RxHigh  = UVMRegField.type_id.create("RxHigh",None, self.get_full_name())
        self.RxFull  = UVMRegField.type_id.create("RxFull",None, self.get_full_name())
        self.SA      = UVMRegField.type_id.create("OOSA",None, self.get_full_name())

        self.TxEmpty.configure(self, 1, 0, "RO", 1, 1, 1, 0, 0)
        self.TxLow.configure(self, 1, 1, "RO", 1, 1, 1, 0, 0)
        self.TxFull.configure(self, 1, 2, "RO", 1, 0, 1, 0, 0)
        self.RxEmpty.configure(self, 1, 4, "RO", 1, 1, 1, 0, 0)
        self.RxHigh.configure(self, 1, 5, "RO", 1, 0, 1, 0, 0)
        self.RxFull.configure(self, 1, 6, "RO", 1, 0, 1, 0, 0)
        self.SA.configure(self, 1, 8, "W1C", 1, 0, 1, 0, 0)

uvm_object_utils(reg_IntSrc)



class reg_IntMask(UVMReg):
    #
    #   UVMRegField TxEmpty
    #   UVMRegField TxLow
    #   UVMRegField TxFull
    #   UVMRegField RxEmpty
    #   UVMRegField RxHigh
    #   UVMRegField RxFull
    #   UVMRegField SA
    #
    def __init__(self, name = "IntMask"):
        super().__init__(name,9,UVM_NO_COVERAGE)
    #   endfunction
    #
    def build(self):
        self.TxEmpty = UVMRegField.type_id.create("TxEmpty",None, self.get_full_name())
        self.TxLow   = UVMRegField.type_id.create("TxLow",None, self.get_full_name())
        self.TxFull  = UVMRegField.type_id.create("TxFull",None, self.get_full_name())
        self.RxEmpty = UVMRegField.type_id.create("RxEmpty",None, self.get_full_name())
        self.RxHigh  = UVMRegField.type_id.create("RxHigh",None, self.get_full_name())
        self.RxFull  = UVMRegField.type_id.create("RxFull",None, self.get_full_name())
        self.SA      = UVMRegField.type_id.create("OOSA",None, self.get_full_name())

        self.TxEmpty.configure(self, 1, 0, "RW", 0, 0, 1, 0, 0)
        self.TxLow.configure(self, 1, 1, "RW", 0, 0, 1, 0, 0)
        self.TxFull.configure(self, 1, 2, "RW", 0, 0, 1, 0, 0)
        self.RxEmpty.configure(self, 1, 4, "RW", 0, 0, 1, 0, 0)
        self.RxHigh.configure(self, 1, 5, "RW", 0, 0, 1, 0, 0)
        self.RxFull.configure(self, 1, 6, "RW", 0, 0, 1, 0, 0)
        self.SA.configure(self, 1, 8, "RW", 0, 0, 1, 0, 0)

uvm_object_utils(reg_IntMask)

#
#
class reg_TxStatus(UVMReg):
    #
    #   UVMRegField TxEn
    #
    def __init__(self, name = "TxStatus"):
        super().__init__(name,1,UVM_NO_COVERAGE)
    #   endfunction
    #
    def build(self):
        self.TxEn = UVMRegField.type_id.create("TxEn",None, self.get_full_name())
        self.TxEn.configure(self, 1, 0, "RW", 0, 0, 1, 0, 0)

uvm_object_utils(reg_TxStatus)


class reg_TxLWM(UVMReg):
    #
    #   UVMRegField TxLWM
    #
    def __init__(self, name = "TxLWM"):
        super().__init__(name,5,UVM_NO_COVERAGE)
    #   endfunction
    #
    def build(self):
        self.TxLWM = UVMRegField.type_id.create("TxLWM",None, self.get_full_name())
        self.TxLWM.configure(self, 5, 0, "RW", 0, 8, 1, 0, 0)

uvm_object_utils(reg_TxLWM)


class reg_RxStatus(UVMReg):
    #
    #   UVMRegField RxEn
    #   UVMRegField Align
    #
    def __init__(self, name = "RxStatus"):
       super().__init__(name,2,UVM_NO_COVERAGE)

    #
    def build(self):
        self.RxEn = UVMRegField.type_id.create("RxEn",None, self.get_full_name())
        self.RxEn.configure(self, 1, 0, "RW", 0, 0, 1, 0, 0)

        self.Align = UVMRegField.type_id.create("Align",None, self.get_full_name())
        self.Align.configure(self, 1, 1, "RO", 0, 0, 1, 0, 0)

uvm_object_utils(reg_RxStatus)


class reg_RxHWM(UVMReg):
    #
    #   UVMRegField RxHWM
    #
    def __init__(self, name = "RxHWM"):
        super().__init__(name,5,UVM_NO_COVERAGE)
    #
    def build(self):
        self.RxHWM = UVMRegField.type_id.create("RxHWM",None, self.get_full_name())
        self.RxHWM.configure(self, 5, 0, "RW", 0, 16, 1, 0, 0)

uvm_object_utils(reg_RxHWM)


class reg_TxRx(UVMReg):
    #
    #   UVMRegField TxRx
    #      
    def __init__(self, name = "TxRx"):
        super().__init__(name,8,UVM_NO_COVERAGE)
    #   endfunction
    #
    def build(self):
        self.TxRx = UVMRegField.type_id.create("TxRx",None, self.get_full_name())
        self.TxRx.configure(self, 8, 0, "RW", 1, 0x00, 1, 0, 0)
    #   endfunction
    #   
uvm_object_utils(reg_TxRx)
    #   
    #endclass


class reg_dut(UVMRegBlock):


    def __init__(self, name = "dut"):
        super().__init__(name, UVM_NO_COVERAGE)
        self.IntSrc = None
        self.IntMask = None
        self.TxStatus = None
        self.TxLWM = None
        self.RxStatus = None
        self.RxHWM = None
        self.TxRx = None


    def build(self):
        self.IntSrc = reg_IntSrc.type_id.create("IntSrc",None,self.get_full_name())
        self.IntMask = reg_IntMask.type_id.create("IntMask",None, self.get_full_name())
        self.TxStatus = reg_TxStatus.type_id.create("TxStatus",None, self.get_full_name())
        self.TxLWM = reg_TxLWM.type_id.create("TxLWM",None, self.get_full_name())
        self.RxStatus = reg_RxStatus.type_id.create("RxStatus",None, self.get_full_name())
        self.RxHWM = reg_RxHWM.type_id.create("RxHWM",None, self.get_full_name())
        self.TxRx = reg_TxRx.type_id.create("TxRx",None, self.get_full_name())

        self.IntSrc.configure(self)
        self.IntMask.configure(self)
        self.TxStatus.configure(self)
        self.TxLWM.configure(self)
        self.RxStatus.configure(self)
        self.RxHWM.configure(self)
        self.TxRx.configure(self)

        self.IntSrc.build()
        self.IntMask.build()
        self.TxStatus.build()
        self.TxLWM.build()
        self.RxStatus.build()
        self.RxHWM.build()
        self.TxRx.build()

        # define default map
        self.default_map = self.create_map("default_map", 0x0, 4, UVM_LITTLE_ENDIAN, 1)
        self.default_map.add_reg(self.IntSrc,   0x0000, "RW")
        self.default_map.add_reg(self.IntMask,  0x0004, "RW")
        self.default_map.add_reg(self.TxStatus, 0x0010, "RW")
        self.default_map.add_reg(self.TxLWM,    0x0014, "RW")
        self.default_map.add_reg(self.RxStatus, 0x0020, "RW")
        self.default_map.add_reg(self.RxHWM,    0x0024, "RW")
        self.default_map.add_reg(self.TxRx,     0x0100, "RW")


uvm_object_utils(reg_dut)



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------

from cocotb.triggers import Timer

from .sv import sv
from .uvm_object import UVMObject
from .uvm_event import UVMEvent
from .uvm_object_globals import UVM_BIN, UVM_DEC
from .uvm_globals import uvm_zero_delay


class UVMBarrier(UVMObject):
    """
    The UVMBarrier class provides a multiprocess synchronization mechanism. Note
    that this is only for cocotb-based simulation, not real multiprocessing.
    It enables a set of processes to block until the desired number of processes
    get to the synchronization point, at which time all of the processes are
    released.
    """


    #  // Function: new
    #  //
    #  // Creates a new barrier object.
    #
    def __init__(self, name="", threshold=0):
        super().__init__(name)
        self.m_event = UVMEvent({"barrier_",name})
        self.threshold = threshold
        self.num_waiters = 0
        self.auto_reset = 1
        self.at_threshold = 0



    #  // Task: wait_for
    #  //
    #  // Waits for enough processes to reach the barrier before continuing.
    #  //
    #  // The number of processes to wait for is set by the <set_threshold> method.
    async def wait_for(self):
        if self.at_threshold:
            return

        self.num_waiters += 1

        if self.num_waiters >= self.threshold:
            if not self.auto_reset:
                self.at_threshold=1
            self.m_trigger()
            return

        await self.m_event.wait_trigger()


    #  // Function: reset
    #  //
    #  // Resets the barrier. This sets the waiter count back to zero.
    #  //
    #  // The threshold is unchanged. After reset, the barrier will force processes
    #  // to wait for the threshold again.
    #  //
    #  // If the ~wakeup~ bit is set, any currently waiting processes will
    #  // be activated.
    def reset(self, wakeup=1):
        self.at_threshold = 0
        if self.num_waiters:
            if wakeup:
                self.m_event.trigger()
            else:
                self.m_event.reset()
        self.num_waiters = 0


    #  // Function: set_auto_reset
    #  //
    #  // Determines if the barrier should reset itself after the threshold is
    #  // reached.
    #  //
    #  // The default is on, so when a barrier hits its threshold it will reset, and
    #  // new processes will block until the threshold is reached again.
    #  //
    #  // If auto reset is off, then once the threshold is achieved, new processes
    #  // pass through without being blocked until the barrier is reset.
    def set_auto_reset(self, value=1):
        self.at_threshold = 0
        self.auto_reset = value


    #  // Function: set_threshold
    #  //
    #  // Sets the process threshold.
    #  //
    #  // This determines how many processes must be waiting on the barrier before
    #  // the processes may proceed.
    #  //
    #  // Once the ~threshold~ is reached, all waiting processes are activated.
    #  //
    #  // If ~threshold~ is set to a value less than the number of currently
    #  // waiting processes, then the barrier is reset and waiting processes are
    #  // activated.
    def set_threshold(self, threshold):
        self.threshold = threshold
        if threshold <= self.num_waiters:
            self.reset(1)



    #  // Function: get_threshold
    #  //
    #  // Gets the current threshold setting for the barrier.
    def get_threshold(self):
        return self.threshold


    #  // Function: get_num_waiters
    #  //
    #  // Returns the number of processes currently waiting at the barrier.
    #
    def get_num_waiters(self):
        return self.num_waiters


    #  // Function: cancel
    #  //
    #  // Decrements the waiter count by one. This is used when a process that is
    #  // waiting on the barrier is killed or activated by some other means.
    def cancel(self):
        self.m_event.cancel()
        self.num_waiters = self.m_event.get_num_waiters()


    type_name = "UVMBarrier"

    def create(self, name=""):
        v = UVMBarrier(name)
        return v

    def get_type_name(self):
        return UVMBarrier.type_name


    async def m_trigger(self):
        self.m_event.trigger()
        self.num_waiters = 0
        await uvm_zero_delay()  # self process was last to wait; allow other procs to resume first


    def do_print(self, printer):
        printer.print_field_int("threshold", self.threshold, sv.bits(self.threshold),
                UVM_DEC, ".", "int")
        printer.print_field_int("num_waiters", self.num_waiters, sv.bits(self.num_waiters),
                UVM_DEC, ".", "int")
        printer.print_field_int("at_threshold", self.at_threshold, sv.bits(self.at_threshold),
                UVM_BIN, ".", "bit")
        printer.print_field_int("auto_reset", self.auto_reset, sv.bits(self.auto_reset), UVM_BIN,
                ".", "bit")


    def do_copy(self, rhs):
        b = rhs
        super().do_copy(rhs)
        #    if(!sv.cast(b, rhs)  or  (b is None)) return
        self.threshold = b.threshold
        self.num_waiters = b.num_waiters
        self.at_threshold = b.at_threshold
        self.auto_reset = b.auto_reset
        self.m_event = b.m_event



#// 
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#// 
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#// 
#//        http://www.apache.org/licenses/LICENSE-2.0
#// 
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#// 

from uvm import *
from .vip_sequencer import vip_sequencer
from .vip_driver import vip_driver
from .vip_monitor import vip_monitor


class vip_agent(UVMAgent):


    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.hier_objection = False


    def build_phase(self, phase):
        self.sqr = vip_sequencer.type_id.create("sqr", self)
        self.drv = vip_driver.type_id.create("drv", self)
        self.tx_mon = vip_monitor.type_id.create("tx_mon", self)
        self.rx_mon = vip_monitor.type_id.create("rx_mon", self)

        self.rx_mon.hier_objection = self.hier_objection
        self.tx_mon.hier_objection = self.hier_objection
        self.drv.hier_objection = self.hier_objection

        vif = []
        if not UVMConfigDb.get(self, "", "vif", vif):
            uvm_fatal("VIP/AGT/NOVIF", "No virtual interface specified for self agent instance")
        self.vif = vif[0]
        UVMConfigDb.set(self, "tx_mon", "vif", self.vif.tx_mon)
        UVMConfigDb.set(self, "rx_mon", "vif", self.vif.rx)

    def connect_phase(self, phase):
        self.drv.seq_item_port.connect(self.sqr.seq_item_export)


    async def pre_reset_phase(self, phase):
        if self.hier_objection:
            phase.raise_objection(self, "Resetting agent")
        await self.reset_and_suspend()
        if self.hier_objection:
            print("vip_agent dropping objection")
            phase.drop_objection(self)


    async def reset_and_suspend(self):
        #fork
        await sv.fork_join([
           cocotb.fork(self.drv.reset_and_suspend()),
           cocotb.fork(self.tx_mon.reset_and_suspend()),
           cocotb.fork(self.rx_mon.reset_and_suspend())
        ])
        #join
        self.sqr.stop_sequences()

    async def suspend(self):
        await sv.fork_join([
        # fork
            cocotb.fork(self.drv.suspend()),
            cocotb.fork(self.tx_mon.suspend()),
            cocotb.fork(self.rx_mon.suspend()),
        ])
        # join

    async def resume(self):
        #      fork
        await sv.fork_join([
            cocotb.fork(self.drv.resume()),
            cocotb.fork(self.tx_mon.resume()),
            cocotb.fork(self.rx_mon.resume()),
        ])
        #      join

uvm_component_utils(vip_agent)
